library(lattice)
library(RColorBrewer)
library(mcclust)
library(mcclust.ext)

#### Study 1: Identical GPD parameters and independent observations ####

obs <- read.table("Single Cluster Independent/Number of clusters and hyperparameters.txt", header=T)
pdf("Single Cluster Independent/NumberClusters.pdf")
plot(obs$Number_of_clusters, xlab="Sample", ylab="Number of clusters", type='l')
temp <- min(obs$Number_of_clusters) : max(obs$Number_of_clusters)
freq <- rep(0, length(temp) )
for(i in 1:length(temp))
  freq[i] <- sum(obs$Number_of_clusters == temp[i]) / length(obs$Number_of_clusters)
plot(temp, freq, type='h', xlab="Number of Clusters", ylab="Frequency" )
dev.off()

similarity <- as.matrix( read.table("Single Cluster Independent/Similarity.txt") )
pdf("Single Cluster Independent/Clustering.pdf", height=5, width=5)
par(mai=c(1.0, 1.0, 0.3, 0.3))
levelplot(similarity, cuts=7, col.regions = brewer.pal(n = 8, name="Greens"), 
          xlab="Index of Areal Unit", ylab = "Index of Areal Unit", main="Similarity Matrix S")
plot(1:20, minbinder(similarity)$cl, xlab="Site index", ylab="Cluster index")
plot(1:20, minVI(similarity)$cl, xlab="Site index", ylab="Cluster index")
dev.off()

psi_est <- as.matrix( read.table("Single Cluster Independent/Sigma.txt") )
psi_summary <- matrix(0, 20, 3)
psi_summary[,1]   <- apply( psi_est, 1, mean )
psi_summary[,2:3] <- t( apply( psi_est, 1, quantile, prob = c(0.05, 0.95) ) )

nu_est  <- as.matrix( read.table("Single Cluster Independent/Xi.txt") )
nu_summary  <- matrix(0, 20, 3)
nu_summary[,1]   <- apply( nu_est, 1, mean )
nu_summary[,2:3] <- t( apply( nu_est, 1, quantile, prob = c(0.05, 0.95) ) )

pdf("Single Cluster Independent/ParameterEstimates.pdf")
par(mfrow=c(1,2))
plot(psi_summary[,1], col=4, pch=19, ylim=c(1.5, 2.5),
     xlab="Site", ylab="Parameter estimate", main= "Scale parameter")
points( 1:20, rep(2, 20), col=2, pch=15 )
for(k in 1:20)
  lines( c(k,k), psi_summary[k,2:3], col=8, lwd=1.1 )

plot(nu_summary[,1], col=4, pch=19, ylim=c(-0.1, 0.3),
     xlab="Site", ylab="Parameter estimate", main= "Shape parameter")
points( 1:20, rep(0.1, 20), col=2, pch=15 )
for(k in 1:20)
  lines( c(k,k), nu_summary[k,2:3], col=8, lwd=1.1 )
dev.off()


#### Study 2: Identical GPD parameters and dependent observations ####

obs <- read.table("Single Cluster Dependent/Number of clusters and hyperparameters.txt", header=T)

pdf("Single Cluster Dependent/NumberClusters.pdf")
plot(obs$Number_of_clusters, xlab="Sample", ylab="Number of clusters", type='l')
temp <- min(obs$Number_of_clusters) : max(obs$Number_of_clusters)
freq <- rep(0, length(temp) )
for(i in 1:length(temp))
  freq[i] <- sum(obs$Number_of_clusters == temp[i]) / length(obs$Number_of_clusters)
plot(temp, freq, type='h', xlab="Number of Clusters", ylab="Frequency" )
dev.off()

similarity <- as.matrix( read.table("Single Cluster Dependent/Similarity.txt") )
pdf("Single Cluster Dependent/Clustering.pdf", height=5, width=5)
par(mai=c(1.0, 1.0, 0.3, 0.3))
levelplot(similarity, cuts=7, col.regions = brewer.pal(n = 8, name="Greens"), 
          xlab="Index of Areal Unit", ylab = "Index of Areal Unit", main="Similarity Matrix S")
plot(1:20, minbinder(similarity)$cl, xlab="Site index", ylab="Cluster index")
plot(1:20, minVI(similarity)$cl, xlab="Site index", ylab="Cluster index")
dev.off()


sigma_est <- read.table("Single Cluster Dependent/Sigma.txt")
xi_est    <- read.table("Single Cluster Dependent/Xi.txt")

sigma_est_summary <- matrix(0, 20, 3)
xi_est_summary    <- matrix(0, 20, 3)

for(k in 1:20){
  sigma_est_summary[k,1]   <- mean( as.numeric(sigma_est[k,]) )
  sigma_est_summary[k,2:3] <- quantile( as.numeric(sigma_est[k,]), c(0.05, 0.95) )
  
  xi_est_summary[k,1]   <- mean( as.numeric(xi_est[k,]) )
  xi_est_summary[k,2:3] <- quantile( as.numeric(xi_est[k,]), c(0.05, 0.95) )
}

pdf("Single Cluster Dependent/ParameterEstimates.pdf")
par(mfrow=c(1,2))
plot(sigma_est_summary[,1], col=4, pch=19, ylim=c(1.5, 2.5),
     xlab="Site", ylab="Parameter estimate", main= "Scale parameter")
points( 1:20, rep(2, 20), col=2, pch=15 )
for(k in 1:20)
  lines( c(k,k), sigma_est_summary[k,2:3], col=8, lwd=1.1 )

plot(xi_est_summary[,1], col=4, pch=19, ylim=c(-0.1, 0.3),
     xlab="Site", ylab="Parameter estimate", main= "Shape parameter")
points( 1:20, rep(0.1, 20), col=2, pch=15 )
for(k in 1:20)
  lines( c(k,k), xi_est_summary[k,2:3], col=8, lwd=1.1 )
dev.off()

#### Study 3: Three clusters and independent observations ####


obs <- read.table("Three Clusters Independent/Number of clusters and hyperparameters.txt", header=T)
pdf("Three Clusters Independent/NumberClusters.pdf")
plot(obs$Number_of_clusters, type='l', ylab="Number of Clusters", xlab="Sample")
temp <- min(obs$Number_of_clusters) : max(obs$Number_of_clusters)
freq <- rep(0, length(temp) )
for(i in 1:length(temp))
  freq[i] <- sum(obs$Number_of_clusters == temp[i]) / length(obs$Number_of_clusters)
plot(temp, freq, type='h', xlab="Number of Clusters", ylab="Frequency" )
dev.off()

num_clust <- which.max(freq) + min(obs$Number_of_clusters) - 1
similarity <- as.matrix( read.table("Three Clusters Independent/Similarity.txt") )

P <- as.matrix( read.table("Three Clusters Independent/P.txt") )
Q <- as.matrix( read.table("Three Clusters Independent/Q.txt") )

P_trans <- P
P_trans[,2]  <- P[,15]
P_trans[2,]  <- P[15,]
P_trans[,15] <- P[,2]
P_trans[15,] <- P[,2]
P_trans[2,2] <- P_trans[15,15] <- Q[2,2]
P_trans[2,15] <- P[2,15]
P_trans[15,2] <- P[15,2]




pdf("Three Clusters Independent/Empirical Chi.pdf", height=5, width=5)
par( mai=c(1.0, 1.0, 0.3, 0.3) )
levelplot( P_trans / Q, cuts=8, col.regions = brewer.pal(9, "Reds"), 
           xlab=list("Index of Areal Unit", cex=1.5), ylab = list("Index of Areal Unit", cex=1.5),
           scales=list(x=list(at=c(5,10,15,20), labels=c(5,10,15,20), cex=1.5), 
                       y = list(at=c(5,10,15,20), labels=c(5,10,15,20)), cex=1.5),
           colorkey = list(labels=list(cex=1.5)) )
dev.off()

similarity_trans <- similarity
similarity_trans[,2] <- similarity[,15]
similarity_trans[2,] <- similarity[15,]
similarity_trans[,15] <- similarity[,2]
similarity_trans[15,] <- similarity[,2]
similarity_trans[2,2] <- similarity_trans[15,15] <- 1
similarity_trans[2,15] <- similarity_trans[15,2] <- similarity[2,15]


pdf("Three Clusters Independent/SimilarityIndependent.pdf", height=5, width=5)
par( mai=c(1.0, 1.0, 0.3, 0.3) )
levelplot( similarity_trans, cuts=8, col.regions = brewer.pal(9, "Greens"), 
           xlab=list("Index of Areal Unit", cex=1.5), ylab = list("Index of Areal Unit", cex=1.5),
           scales=list(x=list(at=c(5,10,15,20), labels=c(5,10,15,20), cex=1.5), 
                       y = list(at=c(5,10,15,20), labels=c(5,10,15,20)), cex=1.5),
           colorkey = list(labels=list(cex=1.5)) )
dev.off()    

pdf("Three Clusters Independent/Clustering.pdf", height=5, width=5)
par(mai=c(1.0, 1.0, 0.3, 0.3))
levelplot(similarity, cuts=7, col.regions = brewer.pal(n = 8, name="Greens"), 
          xlab="Index of Areal Unit", ylab = "Index of Areal Unit", main="Similarity Matrix S")
plot(1:20, minbinder(similarity)$cl, xlab="Site index", ylab="Cluster index")
plot(1:20, minVI(similarity)$cl, xlab="Site index", ylab="Cluster index")
dev.off()


## Parameter estimates
sigma_est <- read.table("Three Clusters Independent/Sigma.txt")
xi_est    <- read.table("Three Clusters Independent/Xi.txt")

## Plot GPD parameter estimates ## 

cluster1 <- c(1, 3, 4, 5, 15)
cluster2 <- c(6, 7, 8, 9, 10, 11, 12, 13, 14)
cluster3 <- c(2, 16, 17, 18, 19, 20)

beta <- c( 2.00, 2.30, 2.60 )
xi   <- c(0.05, 0.10, 0.15)

sigma_true <- xi_true <- rep(0, 20)
sigma_true[cluster1] <- beta[1]
sigma_true[cluster2] <- beta[2]
sigma_true[cluster3] <- beta[3]
xi_true[cluster1] <- xi[1]
xi_true[cluster2] <- xi[2]
xi_true[cluster3] <- xi[3]

quantiles_sigma     <- matrix(0, 20, 3)
quantiles_sigma[,1] <- apply(sigma_est, 1, quantile, prob=0.05)
quantiles_sigma[,2] <- apply(sigma_est, 1, mean)
quantiles_sigma[,3] <- apply(sigma_est, 1, quantile, prob=0.95)

quantiles_xi     <- matrix(0, 20, 3)
quantiles_xi[,1] <- apply(xi_est, 1, quantile, prob=0.05)
quantiles_xi[,2] <- apply(xi_est, 1, mean)
quantiles_xi[,3] <- apply(xi_est, 1, quantile, prob=0.95)

pdf("Three Clusters Independent/GPD estimates.pdf", width=16, height=8)
par(mfrow=c(1,2))
plot(c(), c(), xlim = c(1,20),ylim=c(1.5, 3), xlab="Site", ylab="Parameter", main="Scale parameter" )
for(k in 1:20)
  lines(c(k,k), c(quantiles_sigma[k,1], quantiles_sigma[k,3]) )  
points(1:20, sigma_true, cex=2, pch=15, col=2)
points(1:20, quantiles_sigma[,2], cex=2, pch=16, col=4)
legend("topright", pch=c(15,16,17), col=c(2,4,3), legend=c("Truth", "Posterior Mean", "Individual MLE estimate"))

plot(c(), c(), xlim = c(1,20),ylim=c(-0.1, 0.3), xlab="Site", ylab="Parameter", main="Shape parameter" )
for(k in 1:20)
  lines(c(k,k), c(quantiles_xi[k,1], quantiles_xi[k,3]) )  
points(1:20, xi_true, cex=2, pch=15, col=2)
points(1:20, quantiles_xi[,2], cex=2, pch=16, col=4)
dev.off()


#### Study 4: Three clusters and dependent observations ####


obs <- read.table("Three Clusters Dependent/Number of clusters and hyperparameters.txt", header=T)
pdf("Three Clusters Dependent/NumberClusters.pdf")
plot(obs$Number_of_clusters, type='l', ylab="Number of Clusters", xlab="Sample")
temp <- min(obs$Number_of_clusters) : max(obs$Number_of_clusters)
freq <- rep(0, length(temp) )
for(i in 1:length(temp))
  freq[i] <- sum(obs$Number_of_clusters == temp[i]) / length(obs$Number_of_clusters)
plot(temp, freq, type='h', xlab="Number of Clusters", ylab="Frequency" )
dev.off()

num_clust <- which.max(freq) + min(obs$Number_of_clusters) - 1
similarity <- as.matrix( read.table("Three Clusters Dependent/Similarity.txt") )

pdf("Three Clusters Dependent/Clustering.pdf", height=5, width=5)
par(mai=c(1.0, 1.0, 0.3, 0.3))
levelplot(similarity, cuts=7, col.regions = brewer.pal(n = 8, name="Greens"), 
          xlab="Index of Areal Unit", ylab = "Index of Areal Unit", main="Similarity Matrix S")
plot(1:20, minbinder(similarity)$cl, xlab="Site index", ylab="Cluster index")
plot(1:20, minVI(similarity)$cl, xlab="Site index", ylab="Cluster index")
dev.off()

pdf("Three Clusters Dependent/SimilarityDependent.pdf", height=5, width=5)
par(mai=c(1.0, 1.0, 0.3, 0.3))
levelplot(similarity, cuts=8, col.regions = brewer.pal(n = 9, name="Greens"),
          xlab=list("Index of Areal Unit", cex=1.5), ylab = list("Index of Areal Unit", cex=1.5),
          scales=list(x=list(at=c(5,10,15,20), labels=c(5,10,15,20), cex=1.5), 
                      y = list(at=c(5,10,15,20), labels=c(5,10,15,20)), cex=1.5),
          colorkey = list(labels=list(cex=1.5)) )
dev.off()

## Parameter estimates
sigma_est <- read.table("Three Clusters Dependent/Sigma.txt")
xi_est    <- read.table("Three Clusters Dependent/Xi.txt")

## Plot GPD parameter estimates ## 

cluster1 <- c(1, 3, 4, 5, 15)
cluster2 <- c(6, 7, 8, 9, 10, 11, 12, 13, 14)
cluster3 <- c(2, 16, 17, 18, 19, 20)

beta <- c( 2.00, 2.30, 2.60 )
xi   <- c( 0.05, 0.10, 0.15 )

sigma_true <- xi_true <- rep(0, 20)
sigma_true[cluster1] <- beta[1]
sigma_true[cluster2] <- beta[2]
sigma_true[cluster3] <- beta[3]
xi_true[cluster1] <- xi[1]
xi_true[cluster2] <- xi[2]
xi_true[cluster3] <- xi[3]

quantiles_sigma     <- matrix(0, 20, 3)
quantiles_sigma[,1] <- apply(sigma_est, 1, quantile, prob=0.05)
quantiles_sigma[,2] <- apply(sigma_est, 1, mean)
quantiles_sigma[,3] <- apply(sigma_est, 1, quantile, prob=0.95)

quantiles_xi     <- matrix(0, 20, 3)
quantiles_xi[,1] <- apply(xi_est, 1, quantile, prob=0.05)
quantiles_xi[,2] <- apply(xi_est, 1, mean)
quantiles_xi[,3] <- apply(xi_est, 1, quantile, prob=0.95)

pdf("Three Clusters Dependent/GPD estimates.pdf", width=16, height=8)
par(mfrow=c(1,2))
plot(c(), c(), xlim = c(1,20),ylim=c(1.5, 3), xlab="Site", ylab="Parameter", main="Scale parameter" )
for(k in 1:20)
  lines(c(k,k), c(quantiles_sigma[k,1], quantiles_sigma[k,3]) )  
points(1:20, sigma_true, cex=2, pch=15, col=2)
points(1:20, quantiles_sigma[,2], cex=2, pch=16, col=4)
legend("topright", pch=c(15,16,17), col=c(2,4,3), legend=c("Truth", "Posterior Mean", "Individual MLE estimate"))

plot(c(), c(), xlim = c(1,20),ylim=c(-0.1, 0.3), xlab="Site", ylab="Parameter", main="Shape parameter" )
for(k in 1:20)
  lines(c(k,k), c(quantiles_xi[k,1], quantiles_xi[k,3]) )  
points(1:20, xi_true, cex=2, pch=15, col=2)
points(1:20, quantiles_xi[,2], cex=2, pch=16, col=4)
dev.off()