library(fExtremes)
library(RColorBrewer)
library(lattice)
library(mvtnorm)


filenames <- c("Single Cluster Dependent/","Single Cluster Independent/", 
               "Three Clusters Independent/", "Three Clusters Dependent/")

#### First Test - Single cluster with independent observations ####

set.seed(1000)

n   <- 100
psi <- 2.0
nu  <- 0.1
obs <- matrix( rgpd(20 * n, xi=nu, beta=psi), ncol = n )

Q <- matrix( 20, 20, 20 )
P <- matrix(  0, 20, 20 )

gamma0 <- 2
beta   <- 10

for(k in 1:20){
  for(j in k:20){
    alpha  <- beta / ( exp( gamma0 * dist_municip[k,j] ) - 1 )  
    chi    <- rbeta(1, alpha, beta ) 
    P[k,j] <- P[j,k] <- rbinom(1, 20, chi)
  }
}

write.table(obs, "Single Cluster Independent/PeaksOverThreshold.txt",
            sep=" ", col.names=F, row.names=F)
write.table(P, "Single Cluster Independent/P.txt",
            sep=" ", col.names=F, row.names=F)
write.table(Q, "Single Cluster Independent/Q.txt", 
            sep=" ", col.names=F, row.names=F)

#### First Test - Single cluster with dependent observations ####

obs <- t( apply(obs, 1, sort) )
temp <- sample(n, n)
for(k in 1:20)
  obs[k,] <- obs[k,temp]

write.table(obs, "Single Cluster Dependent/PeaksOverThreshold.txt",
            sep=" ", col.names=F, row.names=F)
write.table(P, "Single Cluster Dependent/P.txt",
            sep=" ", col.names=F, row.names=F)
write.table(Q, "Single Cluster Dependent/Q.txt", 
            sep=" ", col.names=F, row.names=F)


#### Three clusters with independent observations ####

set.seed(1000)

cluster1 <- c( 1, 3, 4, 5, 15 )
cluster2 <- c( 6, 7, 8, 9, 10, 11, 12, 13, 14 )
cluster3 <- c( 2, 16, 17, 18, 19, 20 )

clust_labels <- rep(0,20)
clust_labels[cluster1] <- 1
clust_labels[cluster2] <- 2
clust_labels[cluster3] <- 3

psi <- c( 2.00, 2.30, 2.60 )
nu  <- c( 0.05, 0.10, 0.15 )

n <- 100
obs <- matrix(0, 20, n)
for(k in 1:20)
  obs[k,] <- rgpd( n, xi=nu[clust_labels[k]], beta=psi[clust_labels[k]] )
write.table(obs, "Three Clusters Independent/PeaksOverThreshold.txt",
            sep=" ", col.names=F, row.names=F)

Q <- matrix(20, 20, 20)
P <- matrix( 0, 20, 20)
gamma <- c( 3.0, 2.0, 2.0, 2.0 )
beta  <- 10

for(k in 1:20){
  for(j in k:20){
    if( clust_labels[k] == clust_labels[j] )
      alpha <- beta / ( exp( gamma[clust_labels[k]+1] * dist_municip[k,j] ) - 1 )
    else
      alpha <- beta / ( exp( gamma[1] * dist_municip[k,j] ) - 1 )
    chi <- rbeta(1, alpha, beta) 
    P[k,j] <- P[j,k] <- rbinom(1, 20, chi)
  }
}

write.table(P, "Three Clusters Independent/P.txt",
            sep=" ", col.names=F, row.names=F)
write.table(Q, "Three Clusters Independent/Q.txt", 
            sep=" ", col.names=F, row.names=F)

pdf("Three Clusters Independent/Empirical Chi.pdf", height=5, width=5)
par( mai=c(1.0, 1.0, 0.3, 0.3) )
levelplot( P / Q, cuts=8, col.regions = brewer.pal(9, "Reds"), 
           xlab=list("Index of Areal Unit", cex=1.5), 
           ylab = list("Index of Areal Unit", cex=1.5),
           scales=list(x=list(at=c(5,10,15,20), labels=c(5,10,15,20), cex=1.5), 
                       y = list(at=c(5,10,15,20), labels=c(5,10,15,20)), cex=1.5),
           colorkey = list( labels = list(cex = 1.5) )
         )
dev.off()


#### Three clusters with dependent observations ####

set.seed(2000)

R <- matrix(1, 20, 20)
for(k in 1:19){
  for(j in (k+1):20){
    if( clust_labels[k] != clust_labels[j] )
      R[k,j] <- R[j,k] <- 1 - (dist_municip[k,j] ^ 0.5)
    else{ 
      R[k,j] <- R[j,k] <- 1 - (dist_municip[k,j] ^ 0.5) / 2
    }
  }
}
U <- rmvnorm(n , sigma = R, method = "svd")
U <- pnorm(U) 

obs <- matrix(0, 20, n)
psi <- c( 2.00, 2.30, 2.60 )
nu  <- c( 0.05, 0.10, 0.15 )

for(k in 1:20)
  obs[k,] <- qgpd( U[,k], xi=nu[clust_labels[k]], beta=psi[clust_labels[k]] )

write.table( obs, "Three Clusters Dependent/PeaksOverThreshold.txt", sep=" ", col.names=F, row.names=F)
write.table( P, "Three Clusters Dependent/P.txt", sep=" ", col.names=F, row.names=F)
write.table( Q, "Three Clusters Dependent/Q.txt", sep=" ", col.names=F, row.names=F)

correl <- matrix(0, 20, 20)
for(k in 1:20){
  for(j in 1:20){
    correl[k,j] <- cor( obs[k,], obs[j,] )
  }
}


