#include "analysis.hpp"

using namespace std;

/* Initalize output object based on the number of municipalities

   data     - Observations and neighbourhood structure
   results  - Sampled parameters and cluster structure

*/

void initialize_results(const input &data, output &results)
{

  // Initialize vectors to store sampled parameters
  results.sigma = vector<vector<double> > ( data.K, vector<double>() );
  results.xi    = vector<vector<double> > ( data.K, vector<double>() );
  results.eps   = vector<vector<double> > ( data.K, vector<double>() );
  
  // Initalize similarity matrix
  for(int k=0; k<data.K; ++k)
    results.similarity.push_back( vector<int> (data.K, 0) );

}

void update_similarity(const vector<cluster> &Clusters,
		       const vector<int> &labels, output &results)
{

  for(int k=0; k<labels.size(); ++k){
  
    ++results.similarity[k][k];

    // Update frequency of sites k and j being in the same cluster
    for(int j=k+1; j<labels.size(); ++j){

      if( labels[k] == labels[j] ){
	++results.similarity[k][j];
	++results.similarity[j][k];
      }
      
    }

  }
    
}


/* Store sampled parameters and similarity of municipalities 

   INPUT:
   Clusters - Current cluster allocation
   param    - Cluster-independent parameters and hyperparameters
   labels   - Current cluster labels
   results  - Sampled parameters and cluster structure

*/

void store_samples(const vector<cluster> &Clusters, const hyperparam &param,
		   const vector<int> &labels, output &results)
{
  
  // Store sampled number of clusters
  results.num_clusters.push_back( Clusters.size() );

  // Store sampled hyperparametrs
  results.gamma0.push_back( param.gamma0 );
  results.beta.push_back( param.beta );
  results.kappa.push_back( param.kappa );
  results.mu_sigma.push_back( param.mu_sigma );
  results.mu_xi.push_back( param.mu_xi );
  results.theta_sigma.push_back( param.theta_sigma );
  results.theta_xi.push_back( param.theta_xi );
  results.theta_eps.push_back( param.theta_eps );

  //Loop over all sites
  for(int k=0; k<labels.size(); ++k){

    // Update parameter samples for the municipality
    results.sigma[k].push_back( Clusters[labels[k]].sigma );
    results.xi[k].push_back( Clusters[labels[k]].xi );
    results.eps[k].push_back( Clusters[labels[k]].eps );

  }
  
}


/* Write posterior mean of sampled parameters and cluster allocation
   to output files "Parameters.txt" and "Similrity.txt"

   results     - Sampled parameters and cluster structure
   num_samples - Number of samples 
   data        - Observations and neighbourhood structure
   argv        - File directory in which to store output files

*/

void write_to_file(const output &results, const input &data,
		   const RJMCMC &Setup, char *argv[])
{

  int num_samples = results.num_clusters.size();
  
  string filepath = argv[1]; 
  ofstream sigmafile(filepath + "Sigma.txt");
  ofstream xifile(filepath + "Xi.txt");
  ofstream epsfile(filepath + "Epsilon.txt");

  sigmafile << setprecision(5);
  xifile    << setprecision(5);
  epsfile   << setprecision(5);

  // Store sampled parameters for each municipality
  for(int k=0; k<data.K; ++k){
    
    for(int i=0; i<results.sigma[k].size(); ++i){
      sigmafile << results.sigma[k][i] << " ";
      xifile    << results.xi[k][i] << " ";
      epsfile   << results.eps[k][i] << " ";
    }
    
    sigmafile << "\n";
    xifile    << "\n";
    epsfile   << "\n";
    
  }

  // Store simialrity matrix
  ofstream clusterfile(filepath + "Similarity.txt");

  for(int k=0; k<data.K; ++k){
    for(int j=0; j<data.K; ++j){
      clusterfile << ( 1.0 * results.similarity[k][j] ) /
	(Setup.iterations - Setup.burn_in) << " ";
    }
    clusterfile << "\n";
  }
  
  // Store sampled number of clusters and parameters alpha0, beta and kappa
  ofstream hyperfile(filepath + "Number of clusters and hyperparameters.txt");
  hyperfile << setprecision(6);
  hyperfile << "Number_of_clusters " << "Gamma0 "<< "Beta " << "Kappa"<<" ";
  hyperfile << "MuSigma " << "MuXi ";
  hyperfile << "ThetaSigma " << "ThetaXi "<<"ThetaEps "<< "\n";

  for(int j=0; j<results.num_clusters.size(); ++j){
    
    hyperfile << results.num_clusters[j] << " ";
    hyperfile << results.gamma0[j] << " ";
    hyperfile << results.beta[j] << " ";
    hyperfile << results.kappa[j] << " ";
    hyperfile << results.mu_sigma[j] << " ";
    hyperfile << results.mu_xi[j] << " ";
    hyperfile << results.theta_sigma[j] << " ";
    hyperfile << results.theta_xi[j] << " ";
    hyperfile << results.theta_eps[j] << "\n";
    
  }
  
}

