#include "densities.hpp"

using namespace std;


/* Log-density of a beta-binomial distribution

   Input:
   k - Number of successes
   n - Number of trials 
   a - First parameter
   b - Second parameter
 
*/

double log_dBetaBinom( const int &k, const int &n,
		       const double &a, const double &b )
{
  return( lgamma(n+1.0)   - lgamma(k+1.0) - lgamma(n-k+1.0) -
	  lgamma(n+a+b)   + lgamma(k+a)   + lgamma(n-k+b) +
	  lgamma(a+b)     - lgamma(a)     - lgamma(b) );
}


/* Log-density of a Gamma distribution

   Input:
   x - Value at which density is evaluated
   a - Shape parameter  
   b - Rate parameter 

*/

double log_dGamma(const double &x, const double &a, const double &b)
{
  return( a * log(b) - lgamma(a) + (a-1.0) * log(x) - b * x );   
}


/* Log-density of a Normal density

   Input:
   x - Value at which density is evaluated
   m - Mean
   v - Variance

*/

double log_dNormal(const double &x, const double &m, const double &v)
{
  return( -0.5 * log(2.0 * M_PI * v) - 0.5 / v * pow(x-m, 2.0) );
}


/* Log-density of a Log-normal distribution  
  
   Input:
   x - Value at which density is evaluated
   m - Mean parameter
   v - Variance parameter

*/

double log_dlogNormal(const double &x, const double &m, const double &v)
{ 
  return( -log(x) - 0.5 * log(2.0 * M_PI * v) - 0.5 / v * pow(log(x)-m, 2.0) );
}


/* Log-likelihood function for generalized Pareto distribution

   INPUT:
    pot     - Observed threshold exceedances
    max_pot - Highest observed peak-over threshold
    sigma   - Scale parameter
    xi      - Shape parameter 

*/

double GPD_llike(const vector<double> &pot, const double &max_pot,
		 const double &sigma, const double &xi)
{

  // Check if the peaks-over threshold are feasible for the given sigma and xi
  if( xi < 0.0 && max_pot > -sigma / xi )
    return( -1.0e+100 );

  // Calculate the log-likelihood
  double llike = 0.0;
  for(int i=0; i<pot.size(); ++i)
    llike = llike - log(sigma) - (1.0 / xi + 1.0) * log( 1.0 + xi * pot[i] / sigma );
  
  return( llike );
  
}


/* Gradient of the GPD log-likelihood function

   INPUT:
   pot   - Observed peaks-over threshold
   sigma - Scale parameter
   xi    - Shape parameter

*/

vector<double> GPD_llike_grad(const vector<double> &pot,
			      const double &sigma, const double &xi)
{
  
  vector<double> grad(2, 0.0);
  
  for(int i=0; i<pot.size(); ++i){

    // Gradient with respect to the scale parameter sigma
    grad[0] = grad[0] + ( pot[i] - sigma ) / ( sigma * ( sigma + xi * pot[i] ) ); 

    // Gradient with respect to the shape parameter xi
    grad[1] = grad[1] + 1.0 / pow(xi, 2.0) * log( 1.0 + xi * pot[i] / sigma ) -
      ( 1.0 + 1.0 / xi ) * pot[i] / ( sigma + xi * pot[i] );
    
  }
    
  return( grad );
  
}


/* Derive the adjusted log-likelihood for a cluster

   INPUT:
   obs     - Peaks over threshold
   Cluster - Parameters describing the cluster
   max_pot - Highest observed peaks-over threshold

*/

double GPD_llike_curv(const vector<double> &pot, const cluster &Cluster,
		      const double &max_pot)
{

  // Check whether observation exceeds upper bound
  if( Cluster.xi < 0.0 && max_pot >= -Cluster.sigma / Cluster.xi )
    return( -1e+100 );

  // Derive adjusted scale and shape parameter
  double sigma = Cluster.MLE[0] +
    Cluster.B[0] * ( Cluster.sigma - Cluster.MLE[0] ) +
    Cluster.B[1] * ( Cluster.xi    - Cluster.MLE[1] );

  // Derive adjusted shape parameter
  double xi = Cluster.MLE[1] +
    Cluster.B[2] * ( Cluster.sigma - Cluster.MLE[0] ) +
    Cluster.B[3] * ( Cluster.xi    - Cluster.MLE[1] );
  
  // Check whether observation exceeds upper bound
  if( xi < 0.0 && max_pot >= -sigma / xi )
    return( -1e+100 );

  // Derive adjusted log-likelihood 
  double llike = 0.0;
  for(int i=0; i<pot.size(); ++i)
    llike = llike - log( sigma ) - ( 1.0 / xi + 1.0 ) * log( 1.0 + xi * pot[i] / sigma );
  
  return( llike );
  
}


/* Derive log-likelihood for data on pairwise extremal dependence across all sites

   INPUT:
   Clusters - Current cluster centres and parameters
   data     - Observations and neighbourhood structure 
   labels   - Current cluster allocation of the sites
   param    - Cluster-independent parameters gamma0 and beta

*/

double Chi_llike(const vector<cluster> &Clusters, const input &data,
		 const vector<int> &labels, const hyperparam &param)
{
  
  double log_like = 0.0;
  double alpha;
  double beta = param.beta;
  
  for(int k=0; k<data.K; ++k){
    
    for(int i=0; i<data.neighbours[k].size(); ++i){

      // Derive index of adjacent site
      int j = data.neighbours[k][i];
      
      // Case 1: Adjacent sites are in the same cluster
      if( labels[k] == labels[j] ){

	// Derive second parameter of the beta-binomial distribution
	alpha = Clusters[labels[k]].gamma * data.dist[k][j];
	alpha = beta / ( exp( alpha ) - 1.0 );

	// Update log-likelihood
	log_like = log_like +
	  0.5 * log_dBetaBinom( data.P[k][j], data.Q[k][j], alpha, beta );

      }
      
      // Case 2: Adjacent sites are in different clusters
      else{

	// Derive second parameter of the beta-binomial distribution
	alpha = beta / ( exp( param.gamma0 * data.dist[k][j] ) - 1.0 );

	// Update log-likelihood
	log_like = log_like +
	  0.5 * log_dBetaBinom( data.P[k][j], data.Q[k][j], alpha, beta );
		
      }

    }
    
  }

  return( log_like );
 
}


/* Derive log-likelihood of Chi measure across a single cluster

   Cluster - Current cluster centre and parameters
   data    - Observations and neighbourhood structure 
   labels  - Current cluster allocation of the sites
   param   - Cluster-independent parameters gamma0 and beta

*/

double Chi_llike_Cluster(const cluster &Cluster, const input &data,
			 const vector<int> &labels, const hyperparam &param)
{
  
  double log_like = 0.0;
  double alpha;
  double beta = param.beta;

  // Loop over al sites within the cluster
  for(int i=0; i<Cluster.sites.size(); ++i){

    // Derive site index
    int k1 = Cluster.sites[i];
  
    // Loop over the neighbours of site k
    for(int j=0; j<data.neighbours[k1].size(); ++j){

      // Derive site index
      int k2 = data.neighbours[k1][j];
      
      // Check whether sites are in the same cluster
      if( labels[k1] == labels[k2] ){

	// Derive second parameter of the beta-binomial distribution
	alpha = Cluster.gamma * data.dist[k1][k2];
	alpha = beta / ( exp( alpha ) - 1.0 );

	// Update log-likelihood function
	log_like = log_like +
	  0.5 * log_dBetaBinom( data.P[k1][k2], data.Q[k1][k2], alpha, beta );
	
      }

      // No update required if sites in different clusters
      else
	continue;
	  
    }
    
  }
  
  return( log_like );
 
}
