#include "estimate_eta.hpp"

using namespace std;



/* Derive matrix B for curvature adjustment

   INPUT:
    Cluster - Cluster for which matrix B is to be computed
    data    - Peaks-over threshold and neighbourhood structure 

   OUTPUT:
    Matrix B for curvature adjustment

*/

void derive_B( cluster &Cluster, const input &data )
{
  
  // Merge peaks-over threshold (POT) for sites within cluster
  vector<double> pot;
  double max_pot = 0.0;
  int k;
  
  for(int i=0; i<Cluster.sites.size(); ++i){
    k = Cluster.sites[i];
    if( data.max_peaks[k] > max_pot )
      max_pot = data.max_peaks[k];
    pot.insert( pot.end(), data.peaks[k].begin(), data.peaks[k].end() );
  }

  // Derive maximum likelhood estimators for the scale and shape parameters
  Cluster.MLE = derive_MLE( pot, max_pot );

  // If cluster contains only one site, set B to the identity matrix
  if( Cluster.sites.size() == 1 ){
    Cluster.B = vector<double> ( {1.0, 0.0, 0.0, 1.0} );
    return;
  }
  
  // Compute observed information matrix
  double H11 = derive_H11( pot, Cluster.MLE );
  double H12 = derive_H12( pot, Cluster.MLE );
  double H22 = derive_H22( pot, Cluster.MLE );

  // Derive gradient of the log-likelihood for different periods
  vector<double> grad_sigma( data.peaks_sep[0].size(), 0.0 );
  vector<double> grad_xi( data.peaks_sep[0].size(), 0.0 );
  vector<double> grad;
  
  for(int i=0; i<Cluster.sites.size(); ++i){
    
    int k = Cluster.sites[i];

    for(int j=0; j<grad_sigma.size(); ++j){
      
      grad = GPD_llike_grad( data.peaks_sep[k][j], Cluster.MLE[0], Cluster.MLE[1] );
      grad_sigma[j] = grad_sigma[j] + grad[0];
      grad_xi[j] = grad_xi[j] + grad[1];
      
    }

  }

  // Derive covariance matrix V similar to Smith (1990)
  double V11 = 0.0;
  double V12 = 0.0;
  double V22 = 0.0;

  for(int j=0; j<grad_xi.size(); ++j){
    V11 = V11 + pow( grad_sigma[j], 2.0 );
    V12 = V12 + grad_sigma[j] * grad_xi[j];
    V22 = V22 + pow( grad_xi[j], 2.0 );
  }
  V11 = V11 * grad_xi.size() / (1.0 * grad_xi.size() - 1.0);
  V12 = V12 * grad_xi.size() / (1.0 * grad_xi.size() - 1.0);
  V22 = V22 * grad_xi.size() / (1.0 * grad_xi.size() - 1.0);
  

  // Compute matrix HV^{-1}H
  double HVinvH11, HVinvH12, HVinvH21, HVinvH22, det_V;
  HVinvH11 = H11 * (V22 * H11 - V12 * H12) + H12 * (V11 * H12 - V12 * H11);
  HVinvH12 = H11 * (V22 * H12 - V12 * H22) + H12 * (V11 * H22 - V12 * H12);
  HVinvH21 = H12 * (V22 * H11 - V12 * H12) + H22 * (V11 * H12 - V12 * H11);
  HVinvH22 = H12 * (V22 * H12 - V12 * H22) + H22 * (V11 * H22 - V12 * H12);

  det_V = V11 * V22 - V12 * V12;
  HVinvH11 = HVinvH11 / det_V;
  HVinvH12 = HVinvH12 / det_V;
  HVinvH21 = HVinvH21 / det_V;
  HVinvH22 = HVinvH22 / det_V;

  // Derive root of HV^{-1}H
  double MA11, MA12, MA21, MA22, tr_HVinvH, sq_det_HVinvH;
  tr_HVinvH = HVinvH11 + HVinvH22; 
  sq_det_HVinvH = sqrt( HVinvH11 * HVinvH22 - HVinvH12 * HVinvH21 );

  MA11 = (HVinvH11 + sq_det_HVinvH) / ( sqrt(tr_HVinvH + 2 * sq_det_HVinvH) );
  MA12 = HVinvH12 / ( sqrt(tr_HVinvH + 2 * sq_det_HVinvH) );
  MA21 = HVinvH21 / ( sqrt(tr_HVinvH + 2 * sq_det_HVinvH) );
  MA22 = (HVinvH22 + sq_det_HVinvH) / ( sqrt(tr_HVinvH + 2 * sq_det_HVinvH) );
  
  // Derive root of H
  double tr_H = H11 + H22; 
  double sq_det_H = sqrt( H11 * H22 - H12 * H12 );
  double M11 = (H11 + sq_det_H) / ( sqrt(tr_H + 2*sq_det_H) );
  double M12 = H12 / ( sqrt(tr_H + 2 * sq_det_H) );
  double M21 = H12 / ( sqrt(tr_H + 2 * sq_det_H) );
  double M22 = (H22 + sq_det_H) / ( sqrt(tr_H + 2*sq_det_H) );
  
  // Derive matrix B
  double det_M = M11 * M22 - M12 * M21;
  Cluster.B[0] = ( M22 * MA11 - M12 * MA21 ) / det_M;
  Cluster.B[1] = ( M22 * MA12 - M12 * MA22 ) / det_M;
  Cluster.B[2] = ( M11 * MA21 - M21 * MA11 ) / det_M;
  Cluster.B[3] = ( M11 * MA22 - M21 * MA12 ) / det_M;
  
}


/* Compute the maximum-likelihood estimators for GPD distirbution

   INPUT:
    pot     - Observed threshold exceedances
    max_pot - Highest observed peak-over threshold

*/

vector<double> derive_MLE(const vector<double> &pot, const double &max_pot)
{

  // Find starting point based on moment-estimators 
  vector<double> par = M_Estimator( pot );
  double sigma, xi, llike;
  sigma = par[0];
  xi = par[1];

  // Adjust shape parameter until feasible starting point found
  while( GPD_llike( pot, max_pot, sigma, xi ) < -1e+99 )
    xi = xi + 0.02;
  llike = GPD_llike( pot, max_pot, sigma, xi );

  // Derive gradient of log-likelihood function at starting point
  vector<double> grad;
  grad = GPD_llike_grad( pot, sigma, xi );

  // Initialize parameters for the gradient descent optimization
  double step = 1e-5;
  double stop_crit = 1e-8;
  int iter = 1;
  int max_iter = 10000;
  double sigma_p, xi_p, llike_p;
  vector<double> grad_p;

  // Perform gradient descent
  while( iter < max_iter ){
    
    // Propose parameters via gradient descent and compute negative log-likelihood
    sigma_p = sigma + step * grad[0];
    xi_p    = xi    + step * grad[1];
    llike_p = GPD_llike( pot, max_pot, sigma_p, xi_p );

    // If difference in the log-likelihoods is small, stop
    if( abs( llike_p - llike ) < stop_crit )
      return vector<double> ( {sigma_p, xi_p} );
    
    // If parameter setting is infeasible, decrease step size
    else if( isnan(llike_p) == 1 || llike_p < -1e+99 || llike_p < llike ){
      step = step * 0.5;
      ++iter;
    }

    else{

      // Update gradient and step size
      grad_p = GPD_llike_grad( pot, sigma_p, xi_p );
      step =
      	( (sigma_p - sigma) * (grad[0] - grad_p[0]) + (xi_p - xi) * (grad[1] - grad_p[1]) ) /
	( pow( grad_p[0] - grad[0], 2 ) + pow( grad_p[1] - grad[1], 2 ) );
     
      // Update parameters
      sigma = sigma_p;
      xi    = xi_p;      
      llike = llike_p;
      grad  = grad_p;
      ++iter;
	
    }

  }

  // If optimization routine does not converge, stop
  cout << "Algorithm did not converge" << endl;
  cout << GPD_llike( pot, max_pot, par[0], par[1]) <<endl;
  cout << sigma<<" " << xi << " "<< llike<<endl;
  cout << grad[0]<<" " << grad[1] << " " << step << endl;
  cout << pot.size() << endl;
  exit(1);
    
}


/* Derive initial estimates for the shape and scale of
   the GPD based on the empirical mean and variance

   INPUT:
    pot - Observed peaks-over threshold

   OUTPUT:
    Estimated scale and shape parameter

*/

vector<double> M_Estimator( const vector<double> &pot )
{

  // Derive sample mean
  double mean = 0.0;
  for(int i=0; i<pot.size(); ++i){
    mean = mean + pot[i];
  }
  mean = mean / pot.size();

  // Derive sample variance
  double var = 0.0;
  for(int i=0; i<pot.size(); ++i)
    var = var + pow( pot[i] - mean , 2 );
  var = var / ( pot.size() - 1 );

  // Derive estimates
  vector<double> est(2, 0.0);
  est[0] = 0.5 * mean * ( pow(mean,2) / var + 1 );
  est[1] = 0.5 * ( 1 - pow(mean,2) / var );

  return(est);
  
}


/* First diagonal entry of the observed information matrix 
   of a GPD with parameters sigma and xi 

   INPUT:
   pot - Observed oeaks-over threshold
   MLE - Maximum-likelihood estimators for sigma and xi of the GPD

*/ 

double derive_H11(const vector<double> &pot, const vector<double> &MLE)
{
  
  double H11 = 0.0;

  for(int i=0; i<pot.size(); ++i)
    H11 = H11 +
      ( 2.0 * MLE[0] * pot[i] + MLE[1] * pow(pot[i],2.0) - pow(MLE[0],2.0) ) /
      ( pow(MLE[0], 2.0) * pow(MLE[0] + MLE[1] * pot[i], 2.0) );
      
  return(H11);
      
}


/* Off- diagonal entry of the observed information matrix 
   for a GPD with parameters sigma and xi 

   INPUT:
   pot - Observed peaks-over threshold
   MLE - Maximum-likelihood estimators for sigma and xi of the GPD

*/ 

double derive_H12(const vector<double> &pot, const vector<double> &MLE)
{
  
  double H12 = 0.0;

  for(int i=0; i<pot.size(); ++i)
    H12 = H12 + pot[i] / MLE[0] * ( pot[i] - MLE[0] ) /
      pow( MLE[0] + MLE[1] * pot[i], 2.0 );
      
  return(H12);
      
}


/* Second diagonal entry of the observed information matrix
   for a GPD with parameters sigma and xi 

   INPUT:
   pot - Observed oeaks-over threshold
   MLE - Maximum-likelihood estimators for sigma and xi of the GPD

*/ 

double derive_H22(const vector<double> &pot, const vector<double> &MLE)
{
  
  double H22 = 0.0;

  for(int i=0; i<pot.size(); ++i)
    H22 = H22 + 2 / pow(MLE[1], 3) * log(1.0 + MLE[1] * pot[i] / MLE[0]) -
      2.0 / pow(MLE[1], 2.0) * pot[i] / (MLE[0] + MLE[1] * pot[i]) -
      (1.0 + 1.0 / MLE[1]) * pow(pot[i], 2.0) / pow(MLE[0] + MLE[1] * pot[i], 2.0);
      
  return(H22);

}
