#include "input.hpp"

using namespace std;


/* Read settings for the RJMCMC algorithm from "Setup.txt"

   INPUT:
   Setup - Settings for the RJMCMC algorithm (to be read)
   argv  - File path for "Setup.txt"

*/

void read_Setup( RJMCMC &Setup, int argc, char *argv[] )
{
  
  // Check if filepath is provided
  if(argc < 2){
    cout << "Please specify file path to input files" << endl;
    exit(1);
  }

  // Open settings file
  string filepath = argv[1]; 
  ifstream setupfile( filepath + "Setup.txt", ios::in );
  if( setupfile.fail() ){
    cout << "Error: Setup.txt does not exist!" << endl;
    exit(1);
  }

  // Read number of iterations, check that greater than 0
  setupfile.ignore(256, ':');
  setupfile >> Setup.iterations;
  if( Setup.iterations < 1 ){
    cout << "Error: Number of iterations < 1!" << endl;
    exit(1);
  }

  // Read burn-in, check that it is non-negative
  setupfile.ignore(256, ':');
  setupfile >> Setup.burn_in;
  if( Setup.burn_in < 0 ){
    cout << "Error: Burn-in period < 0!" << endl;
    exit(1);
  }

  // Read thinning size, check that it is greater than 0
  setupfile.ignore(256, ':');
  setupfile >> Setup.thinning;
  if( Setup.thinning < 1 ){
    cout << "Thinning size has to be positive!" << endl;
    exit(1);
  }

  // Read proposal probabilities for the different moves
  for(int i=0; i<7; ++i){
    setupfile.ignore(256, ':');
    setupfile >> Setup.prob[i];
  }
  
  // Verify that proposal probabilities are positive and standarize them
  validate_proposal_prob( Setup.prob );

  // Read initial number of clusters and check that it is greater than 0
  setupfile.ignore(256, ':');
  setupfile >> Setup.clusters;
  if( Setup.clusters < 1 ){
    cout << "Initial number of clusters has to be positive!" << endl;
    exit(1);
  }

  // Length of period in which to separate the raw data to estimate the matrix V
  setupfile.ignore(256, ':');
  setupfile >> Setup.cut_off;
  if( Setup.cut_off < 0 )
    cout << "Length of period has to be positive!" << endl;

  // Print RJMCMC Setup in terminal
  cout << endl;
  cout << "RJMCMC CLUSTERING ALGORITHM SETUP:" << endl;
  cout << endl;
  cout << "Number of iterations            " << Setup.iterations << endl;
  cout << "Burn-in period                  " << Setup.burn_in    << endl;
  cout << "Thinning                        " << Setup.thinning   << endl;
  cout << "Probability for Birth           " << Setup.prob[0]    << endl;
  cout << "Probability for Death           " << Setup.prob[1]    << endl;
  cout << "Probability for Shift           " << Setup.prob[2]    << endl;
  cout << "Probability for Update Sigma    " << Setup.prob[3]    << endl;
  cout << "Probability for Update Xi       " << Setup.prob[4]    << endl;
  cout << "Probability for Update Chi      " << Setup.prob[5]    << endl;
  cout << "Probability for Update Hyper    " << Setup.prob[6]    << endl;
  cout << "Initial number of clusters      " << Setup.clusters   << endl;
  cout << endl;
  
}


/* Check probabilities are positive and scale them to add up to 1

 INPUT:
   prob - Vector of user-specified proposal probabilities

*/

void validate_proposal_prob( vector<double> &prob )
{

  // Check that proposal probabilities are positive
  for(int i=0; i<7; ++i){
    if( prob[i] < 0.0 ){
      cout << "Error: Probability for move " << i+1 << "<0!" << endl;
      exit(1);
    }
  }
  
  // Standardize proposal probabilties
  double temp = 0.0;
  for(int i=0; i<7; ++i)
    temp = temp + prob[i];
  for(int i=0; i<7; ++i)
    prob[i] = prob[i] / temp;
  
}


/* Read observed peaks-over threshold and data for extremal dependence

 INPUT:
  data  - Empty data object to be filled
  Setup - Settings for the RJMCMC algorithm
  argv  - File path where data files are stored 

*/

void read_data( input &data, const RJMCMC &Setup, char* argv[] )
{

  /* Read data for peaks-over threshold */ 

  // Open path to file containng peaks-over threshold
  string filepath = argv[1];
  ifstream peaksfile( filepath + "PeaksOverThreshold.txt", ios::in );

  // If file does not exist, stop
  if( peaksfile.fail() ){
    cout << "Error: PeaksOverThreshold.txt does not exist!" << endl;
    exit(1);
  }
  // Otherwise, read observed peaks-over threshold
  else
    read_peaksfile( data, Setup, argv );


  /* Read data for extremal dependence */
  
  // Check if files for P_{k,k'}, Q_{k,k'} and neighbours exist
  ifstream Pfile( filepath + "P.txt", ios::in );
  ifstream Qfile( filepath + "Q.txt", ios::in );
  ifstream neighboursfile( filepath + "Neighbours.txt", ios::in );
  if( Pfile.fail() ){
    cout << "Error: P.txt does not exist!" << endl;
    exit(1);
  }
  else if( Qfile.fail() ){
    cout << "Error: Q.txt does not exist!" << endl;
    exit(1);
  }
  else if( neighboursfile.fail() ){
    cout << "Error: Neighbours.txt does not exist!" << endl;
    exit(1);
  }
  // Only if all files exist, we read the data
  else
    read_neighbours( data, argv );

  // Read or derive pairwise distances between sites
  derive_distances( data, argv );

}


/* Read observations for all sites from "PeaksOverThreshold.txt"

   INPUT:
   data  - Threshold exceedances (to be filled)
   Setup - Settings for the RJMCMC algorithm
   argv  - File path for "PeaksOverThreshold.txt"

*/

void read_peaksfile( input &data, const RJMCMC &Setup, char* argv[] )
{
 
  // Open file path
  string filepath = argv[1];
  ifstream peaksfile( filepath + "PeaksOverThreshold.txt", ios::in );
  string str;
  int k = 0; // Counter for number of sites

  /* Read peaks-over thresholds with each line in the file 
     corresponding to the observations for one site */
  
  while( getline(peaksfile, str) ){
   
    // Add new columns
    data.peaks.push_back( vector<double> () );
    data.peaks_sep.push_back( vector<vector<double> > () );
    data.peaks_sep[k].push_back( vector<double> () );
      
    // Extract line from the file for next site
    istringstream peaks_k( str );
    double peaks_kt;
    double max_peak = 0.0;  // Highest observation for site k
    int j = 1;             // Counter for day within period
    int i = 0;             // Counter for period
    
    // Read each element from the extracted line
    while( peaks_k >> peaks_kt ){

      // Store only positve values since these are the peaks-over threshold
      if( peaks_kt >= 0.0 ){
	data.peaks[k].push_back( peaks_kt );
	data.peaks_sep[k][i].push_back( peaks_kt );
	if(peaks_kt > max_peak)
	  max_peak = peaks_kt;
      }
      
      // Update counters
      if(j < Setup.cut_off)
	++j;
      else{
	j = 1;
	++i;
	data.peaks_sep[k].push_back( vector<double> () );
      }

    }

    // Store highest observation for site k
    data.max_peaks.push_back( max_peak );

    // Delte last column as it has different length than the rest 
    data.peaks_sep[k].erase( data.peaks_sep[k].end() - 1 );
    
    // Update counter for number of sites
    ++k;
    
  }

  // Derive total number K of sites
  data.K = data.peaks.size();

}


/* Read neighbour structure/distances and dependence measure

   INPUT:
   data  - Neighbourhood structure/distances (to be filled)
   argv  - File path for input files

*/

void read_neighbours( input &data, char* argv[] )
{

  // Open file paths to read P_{k,k'} and Q_{k,k'}
  string filepath = argv[1];
  ifstream Pfile( filepath + "P.txt", ios::in );
  ifstream Qfile( filepath + "Q.txt", ios::in );
 
  string str;
  int temp;

  // Loop over the K sites
  for(int k=0; k<data.K; ++k){

    // Read values for P_{k,}
    data.P.push_back( vector<int> () );
    getline(Pfile, str);
    istringstream P_k(str);
    while( P_k >> temp ){
      if(temp < 0){
	cout << "Error: P_{k,k'} has to be non-negative!" << endl;
	exit(1);
      }
      else
	data.P[k].push_back( temp );
    }

    // Read values for Q_{k,}
    data.Q.push_back( vector<int> () );
    getline(Qfile, str);
    istringstream Q_k(str);
    while( Q_k >> temp ){
      if(temp < 1){
	cout << "Error: Q_{k,k'} has to be positive!" << endl;
	exit(1);
      }
      else
	data.Q[k].push_back( temp );
    }
    
  }

  
  /* Read spatial structure of the sites */

  ifstream neighboursfile( filepath + "Neighbours.txt", ios::in );
  
  for(int k=0; k<data.K; ++k){
    
    data.neighbours.push_back( vector<int> () );
    getline(neighboursfile, str);
    istringstream neighbours_k(str);
    while( neighbours_k >> temp )
      data.neighbours[k].push_back( temp - 1 );
    
  }

}


/* Derive or read pairwise distance between sites
   
   INPUT:
   data - Observations and neighbourhood structure/distances
   argv - File path where "Distances.txt" may already be provided

*/

void derive_distances( input &data, char* argv[] )
{
  
  // Read/Derive distances between sites
  string path = argv[1];
  ifstream distancefile( path + "Distances.txt", ios::in );
  data.dist = vector<vector<double> > ( data.K, vector<double>() );

  // If "Distances.txt" does not exist, derive distances 
  if( distancefile.fail() ){

    for(int k=0; k<data.K; ++k){

      int dist = 0;
      bool STOP = false;
      data.dist[k] = vector<double> (data.K, -1.0 );
      data.dist[k][k] = 0.0;
      
      while( STOP == false ){

	STOP = true;

	for(int m=0; m<data.K; ++m){
	  
	  if( abs(data.dist[k][m] - dist) < 0.1 ){

	    STOP = false;
	    
	    for(int n=0; n<data.neighbours[m].size(); ++n){
	      if( data.dist[k][data.neighbours[m][n]] < 0 )
		data.dist[k][data.neighbours[m][n]] = dist + 1;
	    }
	  
	  }
	
	}
      
	++dist;
    
      }
    
    }

  }
    
  // Otherwise, read from "Distances.txt"
  else{
    
    double dist;
    string str;

    for(int k=0; k<data.K; ++k){
    
      getline(distancefile, str);
      istringstream neighbours_k(str);
      while( neighbours_k >> dist )
	data.dist[k].push_back( dist );

    }
      
  }    

}


/* Sample cluster centers and allocate the remaining sites 

   INPUT:
   Clusters   - Vector of Clusters to be filled
   no_centers - Indices of the sites which are not cluster center (to be filled)
   labels     - Vector of cluster labels 
   data       - Peaks-over thresholds and data for extremal dependence
   RNG        - Random number generator
   param      - Hyperparameters of the Bayesian framework

*/

void init_clusters(vector<cluster> &Clusters,
		   vector<int> &no_centers, vector<int> &labels,
		   input &data, mt19937 &RNG, hyperparam &param)
{

  // Set initial kappa to the number of clusters
  param.kappa = Clusters.size();
  labels = vector<int> ( data.K, 0 );
  
  // Start by labeling each site as being no cluster center
  for(int k=0; k<data.K; ++k)
    no_centers.push_back(k);

  // Sample a random permutation of the cluster indices
  shuffle( no_centers.begin(), no_centers.end(), RNG );
  
  /* Given the number of clusters, sample initial cluster centers */

  for(int i=0; i<Clusters.size(); ++i){

    // Derive index of new cluster center
    Clusters[i].center = no_centers[0];
    Clusters[i].sites.push_back( Clusters[i].center );

    // Set cluster label for the cluster center
    labels[Clusters[i].center] = i;

    // Remove index as site becomes center 
    no_centers.erase( no_centers.begin() ); 

    Clusters[i].xi    = 0.5;                       // Set initial parameter xi_j
    Clusters[i].sigma = 5.0;                       // Set initial parameter sigma_j
    Clusters[i].gamma = 2.0;                       // Set initial parameter gamma_j
    Clusters[i].eps   = log( param.gamma0 / 2.0 ); // Set initial parameter eps_j
    
  }

  // Allocate the sites which are not cluster centers
  for(int j=0; j<no_centers.size(); ++j){

    // Derive distance of the site to first cluster center
    double min_dist = data.dist[Clusters[0].center][no_centers[j]];
    int Z_k = 0;

    // Loop over the remaining cluster centers
    for(int i=1; i<Clusters.size(); ++i){
      
      // If distance to Cluster center i+1 smaller than the previous i, update label 
      if( min_dist > data.dist[Clusters[i].center][no_centers[j]] + 1e-10 ){
	Z_k = i;
	min_dist = data.dist[Clusters[i].center][no_centers[j]];
      }
      // Otherwise, do nothing
      else
	continue;
      
    }
    
    // Store cluster label and allocate site to cluster 
    labels[no_centers[j]] = Z_k;
    Clusters[Z_k].sites.push_back( no_centers[j] );

  }
  
  // Derive curvature adjustment for each cluster
  for(int j=0; j<Clusters.size(); ++j)
    derive_B( Clusters[j], data );
  
  // Derive log-likelihood for each site
  for(int k=0; k<data.K; ++k)
    data.llike.push_back( GPD_llike_curv(data.peaks[k], Clusters[labels[k]], data.max_peaks[k] ) );

}
