#include "moves.hpp"

/* Birth move - Proposal to add new cluster center

   INPUT:
   Clusters   - Current clusters
   data       - Observations and neighbourhood structure 
   labels     - Current cluster labels
   no_centers - Sites that are not cluster centers
   param      - Cluster-independent parameters kappa, gamma0 and beta
   Setup      - Settings of the RJMCMC algorithm
   RNG        - Random number generator
   count      - Counter for acceptance rate of Birth

*/

void Birth( vector<cluster> &Clusters, input &data, vector<int> &labels,
	    vector<int> &no_centers, hyperparam &param, const RJMCMC &Setup,
	    mt19937 &RNG, int &count )
{

  // Return if each site is already an individual cluster
  if( Clusters.size() == data.K )
    return;

  // Sample new cluster centre C_new
  shuffle( no_centers.begin(), no_centers.end(), RNG );
  cluster C_new;
  C_new.center = no_centers[0];

  // Sample new cluster label Z_new
  uniform_int_distribution<int> sample_Z( 0, Clusters.size() );
  int Z_new = sample_Z( RNG );

  // Initalize new cluster configuration
  vector<cluster> Clusters_new = Clusters;
  Clusters_new.insert( Clusters_new.begin() + Z_new, C_new );

  // Update cluster labels and derive adjacent clusters
  vector<int> labels_new, Clust_adj;
  labels_new = labels;
  labels_Birth( Clusters_new, Z_new, labels_new, Clust_adj, data );

  // Initialize acceptance probability
  double accept_prob = log( Setup.prob[1] ) - log( Setup.prob[0] ) +
    log( param.kappa ) - log( Clusters_new.size() - 1.0 );

  // Sample parameters and compute prior/proposal ratio 
  hyperparam param_new = param;
  Sample_Param_GPD( Clusters_new, Z_new, Clust_adj, param, accept_prob, RNG);
  Sample_Param_Chi( Clusters_new, Z_new, Clusters, Clust_adj, param_new, accept_prob, RNG);
 
  // Derive likelihood-ratio for GPD component
  vector<double> llike_new = data.llike;
  GPD_Birth( Clusters_new, Z_new, data, Clust_adj,
	     llike_new, accept_prob );
  
  // Derive likelihood-ratio for Chi component
  accept_prob = accept_prob + 
    Chi_llike( Clusters_new, data, labels_new, param_new ) -
    Chi_llike( Clusters, data, labels, param );

  // Update cluster configuration
  uniform_real_distribution<double> standard_uniform(0,1);
  if( accept_prob > log( standard_uniform(RNG) ) ){
    Clusters = Clusters_new;
    param = param_new;
    labels = labels_new;
    data.llike = llike_new;
    no_centers.erase( no_centers.begin() );
    ++count;
  }

}


/* Derive cluster labels when adding cluster at position Z_prop 

   INPUT:
   Clusters  - Proposed cluster structure
   Z_new     - Index of cluster proposed to be added
   labels    - Cluster lables to be updated
   Clust_adj - Clusters affected by the proposed cluster
   data      - Observations and neighbourhood structure 

*/

void labels_Birth( vector<cluster> &Clusters, const int &Z_new, vector<int> &labels,
		   vector<int> &Clust_adj, const input &data )
{

  // Loop over clusters with smaller index than proposed new cluster
  for(int j=0; j<Z_new; ++j){
 
    //Loop over sites currently allocated to the cluster
    for(int i=0; i<Clusters[j].sites.size(); ++i){

      // Derive site index
      int k = Clusters[j].sites[i];
      
      // Check whether site is allocated to the new cluster
      if( data.dist[k][Clusters[Z_new].center] < data.dist[k][Clusters[j].center] ){

	// Allocate site to new cluster and remove from current cluster
	labels[k] = Z_new;
	Clusters[Z_new].sites.push_back( k );
	Clusters[j].sites.erase( Clusters[j].sites.begin() + i );
	--i;
	  
	// Store that the proposed cluster affects Cluster j
	Clust_adj.push_back( j );
	
      }
      
    }
    
  }

  // Loop over clusters with greater index than Z_new
  for(int j=Z_new+1; j<Clusters.size(); ++j){

    //Loop over sites currently allocated to the cluster
    for(int i=0; i<Clusters[j].sites.size(); ++i){

      // Derive site index
      int k = Clusters[j].sites[i];

      // Check whether site is allocated to the new cluster
      if( data.dist[k][Clusters[Z_new].center] <= data.dist[k][Clusters[j].center] ){

	// Allocate site to new cluster and remove from current cluster
	labels[k] = Z_new;
	Clusters[Z_new].sites.push_back( k );
	Clusters[j].sites.erase( Clusters[j].sites.begin() + i );
	--i;
	
	// Store that the proposed cluster affects Cluster j
	Clust_adj.push_back( j );
	
      }

      // Else, site remains in cluster j and we increment its label
      else
	++labels[k];
      
    }

  }
  
}


/* Sample parameters for the proposed cluster
   and compute prior/proposal ratio

   INPUT:
   Clusters    - Cluster configuration after addition of cluster
   Z           - Index at which cluster is added
   Clust_adj   - Indices of the clusters affected by proposal
   param       - Set of hyperparameters 
   accept_prob - Acceptance probability to be updated
   RNG         - Random number generator

*/

void Sample_Param_GPD(vector<cluster> &Clusters, const int &Z_new,
		      const vector<int> &Clust_adj, const hyperparam &param,
		      double &accept_prob, mt19937 &RNG)
{

  // Derive mean of proposal distributions
  double m_sigma = 0.0;
  double m_xi = 0.0;
  
  for(int j=0; j<Clust_adj.size(); ++j){
    m_sigma = m_sigma + Clusters[ Clust_adj[j] ].sigma;
    m_xi    = m_xi    + Clusters[ Clust_adj[j] ].xi;
  }
  m_sigma = m_sigma / Clust_adj.size();
  m_xi    = m_xi    / Clust_adj.size();

  // Reparametrize the proposal for sigma 
  double v_sigma, mu, theta;
  v_sigma = ( exp( param.theta_sigma ) - 1.0 ) *
    exp( 2 * param.mu_sigma + param.theta_sigma );
  mu = log( pow(m_sigma, 2.0) / sqrt( pow(m_sigma, 2.0) + v_sigma ) );  
  theta = log( v_sigma / pow(m_sigma, 2.0) + 1.0 );

  //Sample proposal for sigma and update acceptance probability
  lognormal_distribution<double> sample_sigma( mu, sqrt(theta) );
  Clusters[Z_new].sigma = sample_sigma( RNG );
  accept_prob = accept_prob +
    log_dlogNormal( Clusters[Z_new].sigma, param.mu_sigma, param.theta_sigma ) -
    log_dlogNormal( Clusters[Z_new].sigma, mu, theta );

  // Sample proposal for xi and update acceptance probability
  normal_distribution<double> sample_xi( m_xi, sqrt(param.theta_xi) );
  Clusters[Z_new].xi = sample_xi( RNG );
  accept_prob = accept_prob +
    log_dNormal( Clusters[Z_new].xi, param.mu_xi, param.theta_xi ) -
    log_dNormal( Clusters[Z_new].xi, m_xi, param.theta_xi );

}


/* Sample parameter epsilon for the proposed cluster and compute prior/proposal ratio

   INPUT:
   Clusters    - Cluster configuration after addition of cluster
   Z           - Index at which cluster is added
   Clust_adj   - Indices of the clusters affected by proposal
   param       - Set of hyperparameters 
   accept_prob - Acceptance probability to be updated
   RNG         - Random number generator

*/

void Sample_Param_Chi(vector<cluster> &Clusters_new,  const int &Z,
		      const vector<cluster> &Clusters, const vector<int> &Clust_adj,
		      hyperparam &param_new, double &accept_prob, mt19937 &RNG)
{  

  /* If all sites are in one cluster, we have to sample two proposals */
  if( Clusters.size() == 1 ){

    // Sample proposals for epsilon1 and epsilon2
    exponential_distribution<double> sample_eps( param_new.theta_eps );
    Clusters_new[0].eps = sample_eps( RNG );
    Clusters_new[1].eps = sample_eps( RNG );    

    // Derive new rates 
    param_new.gamma0      = Clusters[0].gamma * exp( Clusters_new[0].eps );
    Clusters_new[0].gamma = Clusters[0].gamma;
    Clusters_new[1].gamma = param_new.gamma0 / exp( Clusters_new[1].eps );
    
    // Compute the acceptance probability
    accept_prob = accept_prob +
      log_dGamma( param_new.gamma0, 1.0, 0.1 ) -
      log_dGamma( Clusters[0].gamma, 1.0, 0.1 ) +
      Clusters_new[0].eps;
						  
  }
  
  /* Otherwise, only one proposal is required */
  else{ 

    // Derive mean of proposal distribution
    double m_eps = 0.0; 
    for(int j=0; j<Clust_adj.size(); ++j)
      m_eps = m_eps + Clusters_new[ Clust_adj[j] ].eps; 
    m_eps = m_eps / Clust_adj.size();
 
    //Sample proposal for epsilon
    exponential_distribution<double> sample_eps( 1.0 / m_eps );
    Clusters_new[Z].eps   = sample_eps( RNG );
    Clusters_new[Z].gamma = param_new.gamma0 / exp( Clusters_new[Z].eps );

    // Update acceptance probability
    accept_prob = accept_prob +
      log( param_new.theta_eps ) - param_new.theta_eps * Clusters_new[Z].eps +
      log( m_eps ) + Clusters_new[Z].eps / m_eps;
    
  }
    
}


/* Derive log-likelihood ratio of the GPD marginals 

   INPUT:
   Clusters    - Proposed cluster configuration
   Z           - Index at which cluster is inserted
   data        - Observations and neighbourhood structure 
   Clust_adj   - Cluster affected by the proposal
   llike       - GPD likelihood for each site
   accept_prob - Acceptance probability to be updated 

*/

void GPD_Birth(vector<cluster> &Clusters, const int &Z,
	       const input &data, vector<int> &Clust_adj,
	       vector<double> &llike, double &accept_prob)
{
  
  // Remove duplicates from the set of clusters which change
  sort( Clust_adj.begin(), Clust_adj.end() );
  Clust_adj.erase( unique( Clust_adj.begin(), Clust_adj.end() ), Clust_adj.end() );
  Clust_adj.push_back( Z );

  /* Derive log-likelihood for sites allocated to the new cluster */

  for(int m=0; m<Clust_adj.size(); ++m){

    // Derive cluster index
    int j = Clust_adj[m];
     
    // Derive matrix B for curvature adjustment
    derive_B( Clusters[j], data );

    // Loop over sites in the cluster
    for(int i=0; i<Clusters[j].sites.size(); ++i){

      // Derive index of the site
      int k = Clusters[j].sites[i];
      
      // Derive log-likelihood in the new cluster
      llike[k] = GPD_llike_curv( data.peaks[k], Clusters[j], data.max_peaks[k] );

      // Update acceptance probability
      accept_prob = accept_prob + llike[k] - data.llike[k];

    }

  }
  
}


/* Death move - Proposal to remove one of the existing clusters

   INPUT:
   Clusters   - Current cluster centers
   data       - Observations and neighbourhood structure 
   labels     - Current cluster allocation of the municipalities
   no_centers - Sites that are not cluster centers
   param      - Cluster-independent parameters kappa, alpha0, beta0
   Setup      - Settings of the RJMCMC algorithm  
   RNG        - Random number generator
   counts     - Counter for acceptance rate of Birth, Death and Shift

*/

void Death(vector<cluster> &Clusters, input &data,
	   vector<int> &labels, vector<int> &no_centers,
	   const hyperparam &param, const RJMCMC &Setup,
	   mt19937 &RNG, int &count )
{

  // If all municipalities are in one cluster, return
  if( Clusters.size() == 1 )
    return;

  // Sample index of cluster to be removed 
  uniform_int_distribution<int> sample_Z( 0, Clusters.size() - 1 );
  int Z_del = sample_Z( RNG );

  // Initalize new clusters strcuture and store removed cluster
  cluster C_del;
  vector<cluster> Clusters_new = Clusters;
  C_del = Clusters[Z_del];
  Clusters_new.erase( Clusters_new.begin() + Z_del );

  // Derive new labels 
  vector<int> labels_new = labels;
  vector<int> Clust_adj;
  labels_Death( Clusters_new, Z_del, C_del, labels_new, Clust_adj, data );
  
  // Prior/proposal ratio of the parameters
  double accept_prob = 
    Ratio_Death( Clusters_new, Clusters, C_del, Clust_adj, param);
  
  // Likelihood ratio for marginal GPD distributions
  vector<double> llike_new = data.llike;
  GPD_Death( Clusters_new, data, Clust_adj, llike_new, accept_prob );

  // Likelihood-ratio for Chi component
  accept_prob = accept_prob +
    Chi_llike( Clusters_new, data, labels_new, param ) - 
    Chi_llike( Clusters, data, labels, param );

  // Add remaining terms to acceptance probability
  accept_prob = accept_prob +
    log( Clusters.size() - 1.0 ) - log( param.kappa ) +
    log( Setup.prob[0] ) - log( Setup.prob[1] );

  // Update cluster configuration
  uniform_real_distribution<double> standard_uniform(0,1);
  if( accept_prob > log( standard_uniform(RNG) ) ){
    Clusters = Clusters_new;
    no_centers.push_back( C_del.center );
    labels = labels_new; 
    data.llike = llike_new;
    ++count;
  }

}


/* Update cluster labels after removing cluster with index Z_del

   INPUT:
   Clusters  - Cluster configuration after removal of cluster
   Z         - Index of removed cluster
   C_del     - Removed cluster 
   labels    - Cluster labels to be updated
   Clust_adj - Clusters which are affected by removal of C_del
   data      - Observations and neighbourhood structure  

*/

void labels_Death(vector<cluster> &Clusters, const int &Z,
		  const cluster &C_del, vector<int> &labels,
		  vector<int> &Clust_adj, const input &data)
{
  
  // Decrement label if current label is greater than Z_del
  for(int j=Z; j<Clusters.size(); ++j)
    for(int i=0; i<Clusters[j].sites.size(); ++i)
      --labels[ Clusters[j].sites[i] ];
  
  // Loop over sites allocated to the cluster to be removed
  for(int i=0; i<C_del.sites.size(); ++i){

    // Derive site index and initalize label
    int k = C_del.sites[i];
    labels[k] = 0;

    // Loop over the remaining clusters
    for(int j=1; j<Clusters.size(); ++j){

      // Update cluster label, if required
      if( data.dist[k][ Clusters[ labels[k] ].center] >
	  data.dist[k][ Clusters[j].center ] + 1e-10 )
	labels[k] = j;
      
    }

    // Store cluster label and allocate site to cluster
    Clust_adj.push_back( labels[k] );
    Clusters[ labels[k] ].sites.push_back( k );
      
  }

}


/* Derive prior/proposal ratio for Death move

   INPUT:
   Clusters_new - Cluster configuration after removal of cluster
   Clusters     - Current cluster configuration
   C_del        - Cluster to be removed
   members      - Indices of the clusters affected by proposal
   param        - Set of hyperparameters 
   accept_prob  - Acceptance probability to b updated

*/

double Ratio_Death(vector<cluster> &Clusters_new,
		   const vector<cluster> &Clusters, const cluster &C_del,
		   const vector<int> &Clust_adj, const hyperparam &param)
{

  double accept_prob = 0.0;

  /* Consider parameters for generalized Pareto distribution */
  
  // Derive mean of proposal distribution for reverse move
  double m_sigma = 0.0;
  double m_xi = 0.0;

  for(int j=0; j<Clust_adj.size(); ++j){
    m_sigma = m_sigma + Clusters_new[ Clust_adj[j] ].sigma;
    m_xi    = m_xi    + Clusters_new[ Clust_adj[j] ].xi;
  }
  m_sigma = m_sigma / Clust_adj.size();
  m_xi    = m_xi    / Clust_adj.size();

  // Reparametrize the proposal for sigma
  double v_sigma, mu, theta;
  v_sigma = ( exp(param.theta_sigma) - 1.0 ) *
    exp( 2.0 * param.mu_sigma + param.theta_sigma );
  mu = log( pow(m_sigma, 2) / sqrt( pow(m_sigma, 2) + v_sigma) );  
  theta = log( v_sigma / pow(m_sigma, 2) + 1.0 );

  // Update acceptance probability
  accept_prob = accept_prob +
    log_dlogNormal( C_del.sigma, mu, theta ) - 
    log_dlogNormal( C_del.sigma, param.mu_sigma, param.theta_sigma ) +
    log_dNormal( C_del.xi, m_xi, param.theta_xi ) - 
    log_dNormal( C_del.xi, param.mu_xi, param.theta_xi );


  /* Consider parameters for extremal dependence */
  
  // Case 1: Only 1 cluster remains
  if( Clusters_new.size() == 1 ){

    // Derive required proposal values for reverse birth move
    Clusters_new[0].gamma = Clusters[0].gamma;
  
    // Update acceptance probability
    accept_prob = accept_prob +
      log_dGamma( Clusters_new[0].gamma, 1, 0.001 ) -
      log_dGamma( param.gamma0, 1, 0.001 ) -
      Clusters[0].eps;
    
  }

  // Case 2: At least 2 clusters remain
  else{

    // Derive mean of proposal for reverse move
    double m_eps = 0.0;
    for(int k=0; k<Clust_adj.size(); ++k)
      m_eps = m_eps + Clusters_new[ Clust_adj[k] ].eps;
    m_eps = m_eps / Clust_adj.size();

    // Update acceptance probability
    accept_prob = accept_prob -
      log( m_eps ) - C_del.eps / m_eps - 
      log( param.theta_eps ) + param.theta_eps * C_del.eps;
    
  }

  return( accept_prob ); 
  
}


/* Derive log-likelihood ratio of the GPD marginals for Death move

   INPUT:
   Clusters    - Cluster configuration after removal of cluster
   data        - Observations and neighbourhood structure 
   Clust_adj   - Clusters affected by the proposal
   llike       - GPD likelihood for each site
   accept_prob - Acceptance probability to be updated 

*/

void GPD_Death(vector<cluster> &Clusters, const input &data,
	       vector<int> &Clust_adj, vector<double> &llike,
	       double &accept_prob)
{

  // Remove duplicates from the set of clusters which change
  sort( Clust_adj.begin(), Clust_adj.end() );
  Clust_adj.erase( unique( Clust_adj.begin(), Clust_adj.end() ), Clust_adj.end() );
  
  /* Derive log-likelihood ratio */

  // Loop over clusters which change
  for(int m=0; m<Clust_adj.size(); ++m){

    // Derive new curvature adjustment for the cluster
    int j = Clust_adj[m];
    derive_B( Clusters[j], data );
 
    // Update acceptance probability
    for(int i=0; i<Clusters[j].sites.size(); ++i){

      // Derive site index
      int k = Clusters[j].sites[i];

      // Derive adjusted log-likelihood
      llike[k] =
	GPD_llike_curv( data.peaks[k], Clusters[j], data.max_peaks[k] );

      // Update acceptance probability
      accept_prob = accept_prob + llike[k] - data.llike[k];

    }
    
  }

}


/* Shift move - Proposal to shift new cluster center to an adjacent
   site that is not cluster center

   INPUT:
   Clusters   - Current cluster centers
   data       - Observations and neighbourhood structure 
   labels     - Current cluster allocation of the municipalities
   no_centers - Sites that are not cluster centers 
   param      - Cluster-independent parameters kappa, alpha0, beta0
   RNG        - Random number generator
   counts     - Counter for acceptance rate of 'Birth', 'Death' and 'Shift'

*/

void Shift(vector<cluster> &Clusters, input &data,
	   vector<int> &labels, vector<int> &no_centers,
	   const hyperparam &param, mt19937 &RNG, int &count)
{

  // Sample cluster center to be shifted
  uniform_int_distribution<int> sample_Z( 0, Clusters.size() - 1 );
  int Z_shift = sample_Z( RNG );

  // Derive potential site indices for new cluster centre
  vector<int> candid;
  candid = data.neighbours[ Clusters[Z_shift].center ];
  remove_centres( candid, no_centers );

  // Check if centre can be shifted
  if( candid.size() == 0 ) 
    return;

  // Sample new centre for cluster with index Z_shift
  vector<cluster> Clusters_new = Clusters;
  uniform_int_distribution<int> sample_centre( 0, candid.size()-1 );
  Clusters_new[Z_shift].center = candid[ sample_centre( RNG ) ];

  // Update cluster labels
  vector<int> labels_new = labels;
  vector<int> Clust_adj;
  labels_Shift( Clusters_new, Z_shift, labels_new, Clust_adj, data );
    
  // Log-likelihood ratio of the GPD component
  double accept_prob = 0.0;
  vector<double> llike_new = data.llike;
  GPD_Shift( Clusters_new, data, Clust_adj, llike_new, accept_prob );

  // Log-likelihood ratio for Chi measure
  accept_prob = accept_prob +
    Chi_llike( Clusters_new, data, labels_new, param ) - 
    Chi_llike( Clusters, data, labels, param );

  // Derive reverse proposal probability
  vector<int> candid_rev;
  int pos = find(no_centers.begin(), no_centers.end(),
		 Clusters_new[Z_shift].center) - no_centers.begin();
  no_centers[pos] =  Clusters[Z_shift].center;
  candid_rev = data.neighbours[ Clusters_new[Z_shift].center ];
  remove_centres( candid_rev, no_centers );
  
  // Update acceptance probability
  accept_prob = accept_prob + log( candid.size() ) - log( candid_rev.size() );
  
  // Update cluster configuration
  uniform_real_distribution<double> standard_uniform(0,1);
  if( accept_prob > log( standard_uniform(RNG) ) ){
    Clusters = Clusters_new;
    data.llike = llike_new;
    labels = labels_new;
    ++count;
  }
  else
    no_centers[pos] = Clusters_new[Z_shift].center;
  
}


/* Remove indices of current cluster centres 

   INPUT:
   candid     - Set of potential new cluster centers
   no_centers - Set of sites that are not cluster centers

*/

void remove_centres( vector<int> &candid, const vector<int> &no_centers )
{

  // Loop over the set of potential new cluster centres
  for(int j=0; j<candid.size(); ++j){

    // If the site is not a cluster center, we keep it in the set
    if( find(no_centers.begin(), no_centers.end(), candid[j]) !=
	no_centers.end() )
      continue;
    
    //Otherwise, we have to remove it
    else{
      candid.erase( candid.begin() + j );
      --j;
    }

  }
  
}


/* Update cluster labels after shifting one cluster center

   INPUT:
   Clusters  - Cluster configuration after shifting the cluster   
   Z         - Label of the cluster which is to be shifted
   labels    - Label allocation to be updated
   Clust_adj - Clusters affected by the shift
   data      - Observations and neighbourhood structure  

*/

void labels_Shift(vector<cluster> &Clusters, const int &Z,
		  vector<int> &labels, vector<int> &Clust_adj,
		  const input &data)
{

  double min_dist;
  int temp = Clusters[Z].sites.size();
  
  //Loop over clusters with index smaller than shifted cluster
  for(int j=0; j<Z; ++j){

    // Loop over all sites currently allocated to cluster j
    for(int i=0; i<Clusters[j].sites.size(); ++i){

      // Derive site index
      int k = Clusters[j].sites[i];

      // Check if cluster label has to be updated
      if( data.dist[k][Clusters[Z].center] < 
	  data.dist[k][Clusters[j].center] - 1e-10 ){

	// Update cluster label and allocate site index to new cluster
	labels[k] = Z;
	Clusters[Z].sites.push_back( k );
	Clusters[j].sites.erase( Clusters[j].sites.begin() + i );
	Clust_adj.push_back( j );
	--i;
	
      }
    
    }

  }

  //Loop over clusters with index larger than shifted cluster
  for(int j=Z+1; j<Clusters.size(); ++j){
    
    // Loop over all sites currently allocated to cluster j
    for(int i=0; i<Clusters[j].sites.size(); ++i){
      
      // Derive site index
      int k = Clusters[j].sites[i];
      
      // Check if cluster label has to be updated
      if( data.dist[k][Clusters[Z].center] < 
	  data.dist[k][Clusters[j].center] + 1e-10 ){

	// Update cluster label and allocate site index to new cluster
	labels[k] = Z;
	Clusters[Z].sites.push_back( k );
	Clusters[j].sites.erase( Clusters[j].sites.begin() + i );
	Clust_adj.push_back( j );
	--i;
	
      }
	
    }
    
  }

  // Loop over all sites within the shifted cluster
  for(int i=0; i<temp; ++i){

    // Derive site index
    int k = Clusters[Z].sites[i];
    labels[k] = 0;

    for(int j=1; j<Clusters.size(); ++j){
   
      // Update cluster label, if required
      if( data.dist[ Clusters[labels[k]].center ][k] >
	  data.dist[ Clusters[j].center ][k] + 1e-10 )
	labels[k] = j;
      
    }
    
    // Attach site index to new cluster, if allocation changes
    if( labels[k] != Z ){
      Clusters[labels[k]].sites.push_back( k );
      Clusters[Z].sites.erase( Clusters[Z].sites.begin() + i );
      Clust_adj.push_back( labels[k] );
      --temp;
      --i;
    }

    
  }

  // Add index of shifted cluster to set of clusters affected by proposal
  if( Clust_adj.size() > 0 )
    Clust_adj.push_back( Z );
  
}


/* Derive log-likelihood ratio of the GPD marginals for Shift move

   INPUT:
   Clusters    - Cluster configuration after shift of cluster
   data        - Observations and neighbourhood structure 
   Clust_adj   - Clusters affected by the proposal
   llike       - New GPD likelihood for each site to be updated
   accept_prob - Acceptance probability to be updated 

*/

void GPD_Shift(vector<cluster> &Clusters, const input &data,
	       vector<int> &Clust_adj, vector<double> &llike,
	       double &accept_prob)
{
  
  // Remove duplicates from the set of clusters which change
  sort( Clust_adj.begin(), Clust_adj.end() );
  Clust_adj.erase( unique( Clust_adj.begin(), Clust_adj.end() ), Clust_adj.end() );
    
  // Update acceptance probability and log-likelihood 
  for(int m=0; m<Clust_adj.size(); ++m){

    // Derive cluster index
    int j = Clust_adj[m];

    // Update curvature adjustment for cluster
    derive_B( Clusters[j], data );

    // Loop over all sites within that cluster
    for(int i=0; i<Clusters[j].sites.size(); ++i){

      // Derive site index
      int k = Clusters[j].sites[i];

      // Derive new log-likelihood
      llike[k] =
	GPD_llike_curv( data.peaks[k], Clusters[j], data.max_peaks[k] );

      // Update acceptance probability
      accept_prob = accept_prob + llike[k] - data.llike[k];

    }
    
  }
	
}


/* Update GPD scale parameter sigma for all clusters 

   Input:
   Clusters   - Current cluster centers
   data       - Observations and neighbourhood structure
   param      - Set of hyperparameters for proposal distribution
   RNG        - Random number generator

*/

void Sigma(vector<cluster> &Clusters, input &data,
	   const hyperparam &param, mt19937 &RNG)
{

  uniform_real_distribution<double> standard_uniform(0,1);
  
  for(int j=0; j<Clusters.size(); ++j){

    // Sample proposal
    normal_distribution<double> sample_sigma( log(Clusters[j].sigma), 0.1 );
    cluster C_new = Clusters[j];
    C_new.sigma = exp( sample_sigma( RNG ) );

    // Initalize additional parameters
    vector<double> llike_new = data.llike;
    double accept_prob =
      log_dlogNormal( C_new.sigma, param.mu_sigma, param.theta_sigma ) -
      log_dlogNormal( Clusters[j].sigma, param.mu_sigma, param.theta_sigma ) +
      log_dNormal( log(Clusters[j].sigma), log(C_new.sigma), 0.01 ) - 
      log_dNormal( log(C_new.sigma), log(Clusters[j].sigma), 0.01 );

    // Derive acceptance probabability and log-likelihood for proposal

    for(int i=0; i<Clusters[j].sites.size(); ++i){

      //Derive cluster index
      int k = Clusters[j].sites[i];

      //Derive log-likelihood for propsoal
      llike_new[k] =
	GPD_llike_curv( data.peaks[k], C_new, data.max_peaks[k] );

      // Update acceptance probability 
      accept_prob = accept_prob + llike_new[k] - data.llike[k];

    }

    // Accept or reject proposal and update parameters
    if( accept_prob > log( standard_uniform( RNG ) ) ){
      Clusters[j].sigma = C_new.sigma;
      data.llike = llike_new;
    }

  }

}


/* Update GPD shape parameter xi for all clusters 

   Input:
   Clusters   - Current cluster centers
   data       - Observations and neighbourhood structure
   param      - Set of hyperparameters for proposal distribution
   RNG        - Random number generator

*/

void Xi( vector<cluster> &Clusters, input &data,
	 const hyperparam &param, mt19937 &RNG )
{


  uniform_real_distribution<double> standard_uniform(0,1);
  
  for(int j=0; j<Clusters.size(); ++j){

    // Sample proposal
    normal_distribution<double> sample_xi( Clusters[j].xi, 0.05 );
    cluster C_new = Clusters[j];
    C_new.xi = sample_xi( RNG );

    // Initalize additional parameters
    double accept_prob =
      log_dNormal( C_new.xi, param.mu_xi, param.theta_xi ) -
      log_dNormal( Clusters[j].xi, param.mu_xi, param.theta_xi );  
    vector<double> llike_new = data.llike;
  
    // Derive acceptance probabability and log-likelihood 

    for(int i=0; i<Clusters[j].sites.size(); ++i){

      //Derive cluster index
      int k = Clusters[j].sites[i];

      //Derive log-likelihood for propsoal
      llike_new[k] =
	GPD_llike_curv( data.peaks[k], C_new, data.max_peaks[k] );

      // Update acceptance probability
      accept_prob = accept_prob + llike_new[k] - data.llike[k];

    }

    // Accept or reject proposal and update parameters
    if( accept_prob > log( standard_uniform( RNG ) ) ){
      Clusters[j].xi = C_new.xi;
      data.llike = llike_new;
    }

  }
  
}


/* Update parameters for distribution of Chi for all clusters 

   Input:
   Clusters - Current cluster centers
   data     - Observations and neighbourhood structure 
   labels   - Current cluster allocation of the municipalities
   param    - Cluster-independent parameters kappa, alpha0, beta
   RNG      - Random number generator

*/

void Chi(vector<cluster> &Clusters, const input &data,
	 const vector<int> &labels, hyperparam &param,
	 mt19937 &RNG)
{

  uniform_real_distribution<double> standard_uniform(0,1);
  hyperparam param_new = param;
  
  /* If one cluster, we only update gamma1 */
  
  if( Clusters.size() == 1 ){
    
    normal_distribution<double> sample_gamma1( 0, 0.05 );
    cluster C_new = Clusters[0];
    C_new.gamma = C_new.gamma * exp( sample_gamma1( RNG ) );

    double accept_prob =
      Chi_llike_Cluster( C_new, data, labels, param ) -
      Chi_llike_Cluster( Clusters[0], data, labels, param ) +
      log_dGamma( C_new.gamma, 1.0, 0.01 ) -
      log_dGamma( Clusters[0].gamma, 1.0, 0.01 );

    if( accept_prob > log( standard_uniform( RNG ) ) )
      Clusters[0] = C_new;
    
  }

  
  /* Else update gamma0, epsilon1,... */

  else{

    /* Update parameter gamma0 */

    // Sample proposal
    vector<cluster> Clusters_new = Clusters;
    normal_distribution<double> sample_gamma0( 0, 0.05 );
    param_new.gamma0 = param.gamma0 * exp( sample_gamma0( RNG ) );

    // Derive acceptance probability
    for(int j=0; j<Clusters.size(); ++j)
      Clusters_new[j].gamma = param_new.gamma0 / exp( Clusters[j].eps );

    double accept_prob =
      Chi_llike( Clusters_new, data, labels, param_new ) -
      Chi_llike( Clusters, data, labels, param ) +
      log_dGamma(param_new.gamma0, 1.0, 0.001) - 
      log_dGamma(param.gamma0, 1.0, 0.001);
    
    // Accept or reject proposal
    if( accept_prob > log( standard_uniform( RNG ) ) ){
      param = param_new;
      Clusters = Clusters_new; 
    }


    /* Update epsilon1,.. */

    exponential_distribution<double> sample_eps( param.theta_eps );

    for(int j=0; j<Clusters.size(); ++j){

      // Sample proposal
      cluster C_new       = Clusters[j];
      C_new.eps   = sample_eps( RNG );
      C_new.gamma = param.gamma0 / exp( C_new.eps );
     
      // Derive acceptance probability
      double accept_prob =
	Chi_llike_Cluster( C_new, data, labels, param ) -
	Chi_llike_Cluster( Clusters[j], data, labels, param );
      
      // Accept or reject proposal
      if( accept_prob > log( standard_uniform( RNG ) ) )
	Clusters[j] = C_new;

    }

  }

  /* Update parameter beta */
  
  // Sample proposal
  normal_distribution<double> sample_beta( 0.0, 0.01 );
  param_new = param;
  param_new.beta = param.beta * exp( sample_beta( RNG ) );
 
  // Derive acceptance probability
  double accept_prob = 
    Chi_llike( Clusters, data, labels, param_new ) -
    Chi_llike( Clusters, data, labels, param ) +
    log_dGamma( param_new.beta, 1.0, 0.01 ) -
    log_dGamma( param.beta, 1.0, 0.01 );
  
  // Accept or reject proposal
  if( accept_prob > log( standard_uniform( RNG ) ) )
    param = param_new;
  
}


/* Update the parameters of the prior distribution for sigma, xi and epsilon

 INPUT:
 Clusters - Current cluster configuration 
 param    - Current parameters to be updated
 RNG      - Random number generator

*/

void Hyper(const vector<cluster> &Clusters,
	   hyperparam &param, mt19937 &RNG)
{

  // Update Poisson rate kappa
  gamma_distribution<double> sample_kappa( Clusters.size(), 1 / 1.001 );
  param.kappa = sample_kappa( RNG );


  /* Update prior mean for sigma and xi */

  // Derive mean and variance of the full conditional posteriors
  double m_mu_sigma = 0.0;
  double m_mu_xi    = 0.0;
  double v_mu_sigma, v_mu_xi;

  for(int j=0; j<Clusters.size(); ++j){
    m_mu_sigma = m_mu_sigma + log( Clusters[j].sigma ) / param.theta_sigma;
    m_mu_xi = m_mu_xi + Clusters[j].xi / param.theta_xi;
  }
  
  v_mu_sigma = 1.0 / ( 1.0 / 1.0 + Clusters.size() / param.theta_sigma );
  v_mu_xi = 1.0 / ( 1.0 / 0.2 + Clusters.size() / param.theta_xi );
  m_mu_sigma = m_mu_sigma * v_mu_sigma;
  m_mu_xi = m_mu_xi * v_mu_xi;
   
  // Sample new parameter values from the full conditional normal posteriors 
  normal_distribution<double> sample_mu_sigma( m_mu_sigma, sqrt(v_mu_sigma) );
  param.mu_sigma = sample_mu_sigma( RNG );
  normal_distribution<double> sample_mu_xi( m_mu_xi, sqrt(v_mu_xi) );
  param.mu_xi    = sample_mu_xi( RNG );

  
  /* Update prior variance for sigma and xi*/

  double shape   = 1.0 + Clusters.size() / 2.0;
  double r_sigma = 0.1;
  double r_xi    = 0.1;
 
  for(int j=0; j<Clusters.size(); ++j){
    r_sigma = r_sigma + 0.5 * pow( log( Clusters[j].sigma ) - param.mu_sigma, 2.0 );
    r_xi = r_xi + 0.5 * pow( Clusters[j].xi - param.mu_xi, 2.0 );
  }

  gamma_distribution<double> sample_theta_sigma( shape, 1.0 / r_sigma);
  param.theta_sigma = 1.0 / sample_theta_sigma( RNG );
  gamma_distribution<double> sample_theta_xi( shape, 1.0 / r_xi );
  param.theta_xi = 1.0 / sample_theta_xi( RNG );

  /* Update rate for epsilon */
  
  if( Clusters.size() > 1 ){
  
    shape = 5.0;
    double r_eps = 2.0;

    for(int j=0; j<Clusters.size(); ++j){
      if( Clusters[j].sites.size() > 1 ){
	shape = shape + 1.0;
	r_eps = r_eps + Clusters[j].eps;
      }
    }
  
    gamma_distribution<double> sample_theta_eps( shape, 1.0 / r_eps );
    param.theta_eps = sample_theta_eps( RNG );

  }
  
}
