library(mcclust.ext)
library(sp)
library(fExtremes)
library(ismev, quietly = T)

gadm <- readRDS("NOR_adm2.rds")
load("Mapping.RData")

## Function to compute V measure by Rosenberg and Hirschberg (2007)

xlogx <- function(A){
  A[A>0] <- A[A>0] * log( A[A>0] )
}

v.measure <- function( truth, estim, beta=1 ) {
    
  A <- table(truth,estim)
  N <- sum(A)
    
  HB  <- log(N) - sum(xlogx(rowSums(A)))/N
  HC  <- log(N) - sum(xlogx(colSums(A)))/N
  HBC <- sum( xlogx(colSums(A)) )/N - sum(xlogx(A) ) / N
  HCB <- sum( xlogx(rowSums(A)) )/N - sum(xlogx(A) ) / N
    
  h <- 1 - HBC/HB
  c <- 1 - HCB/HC
  v <- (beta+1)*h*c/(beta*h+c)
  v

}

post_median_J   <- rep(0,100)
V_measure_estim <- rep(0,100)

for(i in 1:100){

  setwd( paste("Run",i,sep="") )
  
  ## Read file with the number of clusters
  samples <- read.table("Number of clusters and hyperparameters.txt", header=T)

  ## Derive posterior median number of clusters
  post_median_J[i] <- quantile( samples$Number_of_clusters, 0.5 ) 
  
  ## Export trace plot for the number or clusters J        
  pdf( paste("Samples of J for Run", i, ".pdf", sep=""), height=4, width=5 )
  par( mai=c(1, 1, 0.3, 0.3 ) )
  plot( samples$Number_of_clusters, xlab="Sample", 
        ylab="Number of clusters", type='l', cex.lab=1.5, cex.axis=1.5)
  dev.off()
  
  ## Derive point estimate Z* using the mcclust.ext package by Wade (2015)  
  similarity <- as.matrix( read.table("Similarity.txt") )
  labels_fit <- minVI(similarity)$cl
  
  ## Plot estimated spatial cluster structure
  sites <- c( 1:41, 79, 80, 81, 82, 84, 92:94, 100:102, 105:107 )
  pdf( "ClustersEstimate.pdf" )
  labels_map        <- rep( NA, 430 )
  labels_map[sites] <- labels_fit 
  cluster_plot      <- labels_map[mapping]
  gadm$clusters     <- cluster_plot
  temp              <- gadm[ which(gadm$clusters>0), ]
  plot(temp, col = rainbow(6)[temp$clusters] ) 
  temp <- gadm[35,]
  temp_polygons1 <- temp@polygons[[1]]@Polygons[[1]]
  temp_polygons1 <- Polygons( srl = list(temp_polygons1),ID =  "Nes2")
  temp_polygons1 <- SpatialPolygons( Srl = list(temp_polygons1) )
  plot(temp_polygons1, add=T, col = rainbow(6)[labels_fit [37]])
  dev.off()
  
  ## Quantify the similarity of the point estimate Z* and the true clustering via the V-measure  
  truth <- read.table("Truth.txt")
  V_measure_estim[i] <- v.measure( as.numeric(truth[1,]), labels_fit )
  
  ## Consider parameter estimates for sigma
  samples_sigma <- as.matrix( read.table("Sigma.txt") )
  pdf( paste("SamplesSigmaRun",i,".pdf", sep=""), width=5, height=4 )
  par( mai=c(1, 1, 0.3, 0.3 ) )
  plot(samples_sigma[1,], type='l', xlab="Samples", ylab="Scale", 
       cex.lab=1.5, cex.axis=1.5)
  dev.off()
  
  ## Consider parameter estimates for xi
  samples_xi <- as.matrix( read.table("Xi.txt") )
  pdf( paste("SamplesXiRun",i,".pdf", sep=""), width=5, height=4 )
  par( mai=c(1, 1, 0.3, 0.3 ) )
  plot(samples_xi[1,], type='l', xlab="Samples", ylab="Shape", 
       cex.lab=1.5, cex.axis=1.5)
  dev.off()
  
  ## Compared sampled and true parmaeter
  truth <- as.matrix( read.table("Truth.txt") )
  pdf( paste("ParameterEstimatesRun",i,".pdf", sep=""), width=11, height=4 )
  par( mfrow=c(1,2), mai=c(1, 1, 0.3, 0.3 ) )  
  
  ## Scale parameter
  plot( 1:55, truth[2,], col=2, pch=19, xlab="Site",
        ylab="Scale", cex.lab=1.5, cex.axis=1.5,
        ylim=c( min(apply(samples_sigma,1,quantile, prob=0.05) ), 
                    max(apply(samples_sigma,1,quantile, prob=0.95) ) ) )
  points(1:55, apply(samples_sigma, 1, mean), col=8, pch=18 )
  for(k in 1:55)
    lines( c(k,k), quantile(samples_sigma[k,], c(0.05,0.95), col=8) ) 
  
  plot( 1:55, truth[3,], col=2, pch=19, xlab="Site",
        ylab="Shape", cex.lab=1.5, cex.axis=1.5,
        ylim=c( min(apply(samples_xi,1,quantile, prob=0.05) ), 
                max(apply(samples_xi,1,quantile, prob=0.95) ) ) )
  points(1:55, apply(samples_xi, 1, mean), col=8, pch=18 )
  for(k in 1:55)
    lines( c(k,k), quantile(samples_xi[k,], c(0.05,0.95), col=8) ) 
  dev.off()
  
  setwd("../")
  
}
  
#### Analyse posterior median and V-measure estimates ####

## Plot histogram of posterior estimate for J
range_med <- sort( unique(post_median_J) )
freq      <- rep(0, length(range_med) )
for( i in 1:length(freq) )
  freq[i] = sum( post_median_J == range_med[i]) / 100

pdf("PosteriorMedianClusters.pdf", width=5, height=4)
par(mai=c(1,1,0.3,0.3))
plot( range_med, c(), ylim=c(0,max(freq)), xlim=c( range_med[1], max(range_med) ), 
      xlab="Posterior Median Number of Clusters", ylab = "Frequency", cex.lab=1.5, cex.axis=1.5  )
for(i in 1:length(freq))
  lines( rep(range_med[i],2), c(0,freq[i]), lwd=2 )
dev.off()

## Plot distribution function for the V-measure across the 100 repetitions
print( mean(V_measure_estim) )
pdf("VMeasureECDF.pdf", width=5, height=4)
par(mai=c(1,1,0.3,0.3))
plot(ecdf(V_measure_estim), xlab="V-Measure", ylab="Empirical CDF", main="",
     cex.lab=1.5, cex.axis=1.5)
dev.off()