library(fExtremes)
library(RColorBrewer)
library(lattice)
library(mvtnorm)
library(ismev)
library(sp)

load("Mapping.RData")
source("Neighbourhood StructureNorway.R")
gadm <- readRDS("NOR_adm2.rds")

## Select regions of Ostfold, Akershus and eastern Buskerud - 55 sites
sites             <- c( 1:41, 79, 80, 81, 82, 84, 92:94, 100:102, 105:107 )

## Extract neighbourhood structure
neighbours <- vector("list", length(sites) )
for( i in 1:length(sites) ){
  temp <- neighbours_true[[ sites[i] ]]
  for(k in temp){
    ind <- which( sites == k )
    neighbours[[i]] <- c(neighbours[[i]], ind)
  }
}


## Derive distance between sites on [0,1] scale
K    <- length(sites)
coord_municip <- matrix(0, K, 2)
for(j in 1:K){
  k <- sites[j]
  ind <- which( mapping == k )
  for(i in ind)
    coord_municip[j,] <- coord_municip[j,] + gadm@polygons[[i]]@labpt
  coord_municip[j,] <- coord_municip[j,] / length(ind)
}
coord_municip[23,] <- gadm[35,]@polygons[[1]]@Polygons[[2]]@labpt
coord_municip[37,] <- gadm[35,]@polygons[[1]]@Polygons[[1]]@labpt

dist <- matrix( 0, K, K )
for(k in 1:K){
  for(j in 1:K){
    phi1    <- coord_municip[k,2]
    phi2    <- coord_municip[j,2]
    lambda1 <- coord_municip[k,1]
    lambda2 <- coord_municip[j,1]
    temp_a <- sin( ( phi1 - phi2 ) *pi / 360 ) ^ 2 + 
      cos(phi1 * pi / 180) * cos(phi2 * pi / 180) * sin( ( lambda1 - lambda2 ) * pi / 360 ) ^ 2
    temp_c <- 2 * atan2( sqrt(temp_a), sqrt(1 - temp_a) )
    dist[k,j] <- 6371 * temp_c
  }
}
dist <- dist / max(dist)

############## Sample clusters ####################
set.seed(2019)
J <- 5
Z <- rep(0, length(sites))

## Model parameters ##
n     <- 100
psi   <- seq( 1.5, 2.5, length.out = J )
nu    <- seq( 0.1, 0.3, length.out = J )
gamma <- c( 2.0, rep(1.0, J) )
beta  <- 10

for(i in 1:100){

  ## Sample cluster centres and derive labels
  centres  <- sample(1:K, J, replace = F)
  for(k in 1:K)
    Z[k] <- which.min( dist[k,centres] )

  ## Plot cluster structure
  pdf( paste("Run",i,"/Clusters.pdf", sep="") )
  Z_map         <- rep( NA, 430 )
  Z_map[sites]  <- Z
  cluster_plot  <- Z_map[mapping]
  gadm$clusters <- cluster_plot
  temp      <- gadm[ which(gadm$clusters>0), ]
  plot(temp, col = rainbow(6)[temp$clusters] ) 
  temp <- gadm[35,]
  temp_polygons1 <- temp@polygons[[1]]@Polygons[[1]]
  temp_polygons1 <- Polygons( srl = list(temp_polygons1),ID =  "Nes2")
  temp_polygons1 <- SpatialPolygons( Srl = list(temp_polygons1) )
  plot(temp_polygons1, add=T, col = rainbow(6)[Z[37]])
  
  for(k in centres)
    points(coord_municip[k,1], coord_municip[k,2], pch=19)
  dev.off()
  
  ## Store distance and neighbourhood structure
  write.table( dist, paste("Run",i,"/Distances.txt",sep=""),
               sep=" ", row.names=F, col.names=F )
  write.table( t(neighbours[[1]]), paste("Run",i,"/Neighbours.txt",sep=""),
                 sep=" ", row.names=F, col.names=F )
  for(k in 2:K)
    write.table( t(neighbours[[k]]), paste("Run",i,"/Neighbours.txt",sep=""), 
                   sep=" ", row.names=F, col.names=F, append=T )

  ## Sample peaks over threshold  
  R <- matrix(1, K, K)
  for(k in 1:(K-1)){
    for(j in (k+1):K){
      if( Z[k] != Z[j] )
        R[k,j] <- R[j,k] <- 1 - ( dist[k,j] / max(dist) ) ^ 0.25
      else{ 
        R[k,j] <- R[j,k] <- 1 - ( dist[k,j] / max(dist) ) ^ 0.25 / 1.5
      }
    }
  }
  U <- rmvnorm(n , sigma = R, method = "svd")
  U <- pnorm(U) 
  
  obs <- matrix(0, K, n)
  for(k in 1:K)
    obs[k,] <- qgpd( U[,k], xi=nu[Z[k]], beta=psi[Z[k]] )
  write.table(obs, paste("Run",i,"/PeaksOverThreshold.txt", sep="" ),
              sep=" ", col.names=F, row.names=F)

  ## Sample extremal dependence measures
  Q <- matrix( 25, K, K )
  P <- matrix(  0, K, K )
  alpha <- 0
  for(k in 1:K){
    for(j in k:K){
      if( Z[k] == Z[j] )
        alpha <- beta / ( exp( gamma[Z[k]+1] * dist[k,j] ) - 1 )
      else
        alpha <- beta / ( exp( gamma[1] * dist[k,j] ) - 1 )
      chi <- rbeta(1, alpha, beta) 
      P[k,j] <- P[j,k] <- rbinom(1, 25, chi)
    }
  }
  write.table(P, paste("Run",i,"/P.txt", sep="" ),
              sep=" ", col.names=F, row.names=F)
  write.table(Q, paste("Run",i,"/Q.txt", sep="" ),
              sep=" ", col.names=F, row.names=F)
  
  ## Store truth
  truth <- matrix(0, 3, K)
  truth[1,] <- Z
  truth[2,] <- psi[Z]
  truth[3,] <- nu[Z]
    
  write.table(truth, paste("Run",i,"/Truth.txt", sep="" ),
              sep=" ", col.names=F, row.names=F)
  
}


