library(lubridate, quietly=T)
library(deldir, quietly=T)
library(sp)

#### Load geographical data and derive distances and neighbourhood structure ####

locations <- read.csv( file = "RiverFlowLocations.csv", header = T )
locations <- locations[-31,] ## Remove site with incorrect hydrological coordinates

## Derive hydrological distances and scale them to [0,1] 
hydro_dist <- matrix(0, 45, 45)
for(k1 in 1:45){
  for(k2 in 1:45){ 
    hydro_dist[k1,k2] <- sqrt( (locations$HydroEast[k1] - locations$HydroEast[k2]   ) ^ 2 +  
                               (locations$HydroNorth[k1] - locations$HydroNorth[k2] ) ^ 2 )
  }
}
hydro_dist <- hydro_dist / max(hydro_dist)
write.table( hydro_dist, "Winter/Distances.txt", col.names=F, row.names=F)
write.table( hydro_dist, "Summer/Distances.txt", col.names=F, row.names=F)

#### Derive Voronoi tesselation and store adjacent cells ####
vtess <- deldir( locations$HydroEast, locations$HydroNorth )
neighbours <- vector("list", 45)
for(k in 1:45){
  neighbours[[k]] <- sort( c( 
    vtess$dirsgs$ind2[ which(vtess$dirsgs$ind1 == k) ],
    vtess$dirsgs$ind1[ which(vtess$dirsgs$ind2 == k) ]
  ) )
}
write.table(t(neighbours[[1]]), "Winter/Neighbours.txt", col.names=F, row.names = F)
write.table(t(neighbours[[1]]), "Summer/Neighbours.txt", col.names=F, row.names = F)
for(k in 2:45){
  write.table(t(neighbours[[k]]), "Winter/Neighbours.txt", col.names=F, row.names = F, append=T)
  write.table(t(neighbours[[k]]), "Summer/Neighbours.txt", col.names=F, row.names = F, append=T)
}


#### Load data set ####

obs       <- read.csv( file = "RiverFlowData.csv"     , header = T )

## Exploratory analysis for Bywell
pdf("EDABywell.pdf", width=5, height=5)
par(mai=c(1,1,0.3,0.3))
temp <- rep(0,365)
sample_size <- rep(0,365)
for(i in 1:length(obs$X23001) ){
  if( is.na(obs$X23001[i]) == T )
    next
  else{
    temp[i%%365+1] <- temp[i%%365+1] + obs$X23001[i]
    sample_size[i%%365+1] <- sample_size[i%%365+1] + 1
  }
}
ma <- function(x,n=14){
  filter(x,rep(1/n,n), sides=2)
}
plot(temp / sample_size, xlab="Day of year", ylab="Average", 
     pch=19, col=8, cex.lab=1.5, cex.main=1.5, cex.axis=1.5, cex=1.5)
ind <- c(351:365, 1:365, 1:20)
lines(1:365, ma(temp[ind]/sample_size[ind])[16:380], lwd=2, col=2)
dev.off()


#### Extract data and derive weekly maximum ####

obs_sites      <- vector("list", 46)
obs_aggregated <- vector("list", 46)

for(k in 1:46){
  
  obs_sites[[k]] <- matrix(0, 63, 364)
  temp <- 1
  
  for( i in 1:23010 ){
    day_of_year <- as.numeric( strftime(as.POSIXlt(obs$Date[i], format="%d/%m/%Y"), format="%j") )
    if(day_of_year < 365)
      obs_sites[[k]][temp, day_of_year] <- as.numeric( obs[i,k+2] )
    if(day_of_year == 365)
      temp = temp + 1
  }

  obs_aggregated[[k]] <- matrix(0, 63, 52)  
  for(i in 1:63){
    for(week_k in 1:52){
      obs_aggregated[[k]][i, week_k] = max( obs_sites[[k]][i, (1+(week_k-1)*7) : (week_k*7)] )
    }
  }

}


#### Split data into seasons ####

obs_winter <- obs_summer <- matrix(0, 0, 46)
for( year in 1980:2011 ){
  temp_winter <- matrix(year, 20, 46)
  temp_summer <- matrix(year, 22, 46)
  
  for(k in 1:46){
    temp_winter[,k] <- obs_aggregated[[k]][year-1948, c(1:12,45:52) ] 
    temp_summer[,k] <- obs_aggregated[[k]][year-1948, 17:38]
  }
  obs_winter <- rbind(obs_winter, temp_winter)
  obs_summer <- rbind(obs_summer, temp_summer)
}

## Remove gauge with incorrect hydrological coordinates and store aggregated data
obs_winter <- obs_winter[,-31]
obs_summer <- obs_summer[,-31]
write.table( t(obs_winter), "Winter/AggregatedFlow.txt", col.names = F, row.names = F )
write.table( t(obs_summer), "Summer/AggregatedFlow.txt", col.names = F, row.names = F )

## Scale data site-wise to mean 0 and variance 1 
obs_winter <- apply( obs_winter, 2, scale )
obs_summer <- apply( obs_summer, 2, scale )


#### Derive values P_{k,k'} and Q_{k,k'} ####
Q_winter <- Q_summer <- matrix(0, 45, 45)
P_winter <- P_summer <- matrix(0, 45, 45)

for(k1 in 1:45){
  for(k2 in 1:45){ 
     
    ## Winter
    Nk1k2 <- which( is.na(obs_winter[,k1]) + is.na(obs_winter[,k2]) == 0 ) 
    Qk1   <- which( obs_winter[Nk1k2,k1] > quantile(obs_winter[,k1], 0.965, na.rm=T) )
    Qk2   <- which( obs_winter[Nk1k2,k2] > quantile(obs_winter[,k2], 0.965, na.rm=T) )
    Q_winter[k1,k2] <- length(Qk1)
    Q_winter[k2,k1] <- length(Qk2)
    P_winter[k1,k2] <- P_winter[k2,k1] <- length( intersect(Qk1, Qk2) )
    
    ## Summer
    Nk1k2 <- which( is.na(obs_summer[,k1]) + is.na(obs_summer[,k2]) == 0 ) 
    Qk1   <- which( obs_summer[Nk1k2,k1] > quantile(obs_summer[,k1], 0.95, na.rm=T) )
    Qk2   <- which( obs_summer[Nk1k2,k2] > quantile(obs_summer[,k2], 0.95, na.rm=T) )
    Q_summer[k1,k2] <- length(Qk1)
    Q_summer[k2,k1] <- length(Qk2)
    P_summer[k1,k2] <- P_summer[k2,k1] <- length( intersect(Qk1, Qk2) )
    
  }
}
write.table( Q_summer, "Summer/Q.txt", col.names=F, row.names=F)
write.table( P_summer, "Summer/P.txt", col.names=F, row.names=F)
write.table( Q_winter, "Winter/Q.txt", col.names=F, row.names=F)
write.table( P_winter, "Winter/P.txt", col.names=F, row.names=F)


#### Derive peaks over threshold and export to data files ####

u_winter <- 2.0
u_summer <- 2.0
obs_winter[is.na(obs_winter)] <- -10
obs_winter <- obs_winter - u_winter
obs_summer[is.na(obs_summer)] <- -10
obs_summer <- obs_summer - u_summer
write.table( t(obs_summer), "Summer/PeaksOverThreshold.txt", col.names=F, row.names=F)
write.table( t(obs_winter), "Winter/PeaksOverThreshold.txt", col.names=F, row.names=F)